import numpy as np
from .utils import Counter, sdiag, timeIt, Identity, validate_type, mkvc
from .data import Data
from .simulation import BaseSimulation
from .objective_function import L2ObjectiveFunction

__all__ = ["L2DataMisfit"]


class BaseDataMisfit(L2ObjectiveFunction):
    """
    BaseDataMisfit

    .. note::
        You should inherit from this class to create your own data misfit
        term.
    """

    def __init__(self, data, simulation, model_map=None, debug=False, counter=None, **kwargs):
        super().__init__(**kwargs)

        self.data = data
        self.simulation = simulation
        self.debug = debug
        self.count = counter

        self.model_map = model_map

    @property
    def model_map(self):
        return getattr(self, "_model_map", None)

    @model_map.setter
    def model_map(self, value):
        if value is None:
            value = Identity()
        self._model_map = value
        self._has_fields = True


    @property
    def data(self):
        """A SimPEG data class containing the observed data.

        Returns
        -------
        SimPEG.data.Data
        """
        return self._data

    @data.setter
    def data(self, value):
        self._data = validate_type("data", value, Data, cast=False)

    @property
    def simulation(self):
        """A SimPEG simulation.

        Returns
        -------
        SimPEG.simulation.BaseSimulation
        """
        return self._simulation

    @simulation.setter
    def simulation(self, value):
        self._simulation = validate_type(
            "simulation", value, BaseSimulation, cast=False
        )

    @property
    def debug(self):
        """Print debugging information.

        Returns
        -------
        bool
        """
        return self._debug

    @debug.setter
    def debug(self, value):
        self._debug = validate_type("debug", value, bool)

    @property
    def counter(self):
        """Set this to a ``SimPEG.utils.Counter`` if you want to count things.

        Returns
        -------
        SimPEG.utils.Counter or None
        """
        return self._counter

    @counter.setter
    def counter(self, value):
        if value is not None:
            value = validate_type("counter", value, Counter, cast=False)

    @property
    def nP(self):
        """
        number of model parameters
        """
        if self._mapping is not None:
            return self.mapping.nP
        elif self.simulation.model is not None:
            return len(self.simulation.model)
        else:
            return "*"

    @property
    def nD(self):
        """
        number of data
        """
        return self.data.nD

    @property
    def shape(self):
        """"""
        return (self.nD, self.nP)

    @property
    def W(self):
        """W
        The data weighting matrix.
        The default is based on the norm of the data plus a noise floor.
        :rtype: scipy.sparse.csr_matrix
        :return: W
        """

        if getattr(self, "_W", None) is None:
            if self.data is None:
                raise Exception(
                    "data with standard deviations must be set before the data "
                    "misfit can be constructed. Please set the data: "
                    "dmis.data = Data(dobs=dobs, relative_error=rel"
                    ", noise_floor=eps)"
                )
            standard_deviation = self.data.standard_deviation
            if standard_deviation is None:
                raise Exception(
                    "data standard deviations must be set before the data misfit "
                    "can be constructed (data.relative_error = 0.05, "
                    "data.noise_floor = 1e-5), alternatively, the W matrix "
                    "can be set directly (dmisfit.W = 1./standard_deviation)"
                )
            if any(standard_deviation <= 0):
                raise Exception(
                    "data.standard_deviation must be strictly positive to construct "
                    "the W matrix. Please set data.relative_error and or "
                    "data.noise_floor."
                )
            self._W = sdiag(1 / (standard_deviation))
        return self._W

    @W.setter
    def W(self, value):
        if isinstance(value, Identity):
            value = np.ones(self.data.nD)
        if len(value.shape) < 2:
            value = sdiag(value)
        assert value.shape == (
            self.data.nD,
            self.data.nD,
        ), "W must have shape ({nD},{nD}), not ({val0}, {val1})".format(
            nD=self.data.nD, val0=value.shape[0], val1=value.shape[1]
        )
        self._W = value

    def residual(self, m, f=None):
        if self.data is None:
            raise Exception("data must be set before a residual can be calculated.")
        return self.simulation.residual(m, self.data.dobs, f=f)


class L2DataMisfit(BaseDataMisfit):
    r"""
    The data misfit with an l_2 norm:

    .. math::

        \mu_\text{data} =
            \frac{1}{2}
            \left|
                \mathbf{W}_d (\mathbf{d}_\text{pred} - \mathbf{d}_\text{obs})
            \right|_2^2
    """

    @timeIt
    def __call__(self, m, f=None):
        "__call__(m, f=None)"

        R = self.W * self.residual(m, f=f)
        return 0.5 * np.vdot(R, R)

    @timeIt
    def deriv(self, m, f=None):
        r"""
        Derivative of the data misfit

        .. math::

            \mathbf{J}^{\top} \mathbf{W}^{\top} \mathbf{W}
            (\mathbf{d} - \mathbf{d}^{obs})

        :param numpy.ndarray m: model
        :param SimPEG.fields.Fields f: fields object
        """

        if f is None:
            f = self.simulation.fields(m)

        return self.simulation.Jtvec(
            m, self.W.T * (self.W * self.residual(m, f=f)), f=f
        )

    @timeIt
    def deriv2(self, m, v, f=None):
        r"""
        Second derivative of the data misfit

        .. math::

            \mathbf{J}^{\top} \mathbf{W}^{\top} \mathbf{W} \mathbf{J}

        :param numpy.ndarray m: model
        :param numpy.ndarray v: vector
        :param SimPEG.fields.Fields f: fields object
        """

        if f is None:
            f = self.simulation.fields(m)

        return self.simulation.Jtvec_approx(
            m, self.W * (self.W * self.simulation.Jvec_approx(m, v, f=f)), f=f
        )

    def getJtJdiag(self, m):
        """
        Evaluate the main diagonal of JtJ
        """
        if getattr(self.simulation, "getJtJdiag", None) is None:
            raise AttributeError(
                    "Simulation does not have a getJtJdiag attribute."
                    + "Cannot form the sensitivity explicitly"
            )

        mapping_deriv = self.model_map.deriv(m)

        if self.model_map is not None:
            m = mapping_deriv @ m

        jtjdiag = self.simulation.getJtJdiag(m, W=self.W)

        if self.model_map is not None:
            jtjdiag = mkvc((sdiag(np.sqrt(jtjdiag)) @ mapping_deriv).power(2).sum(axis=0))

        return jtjdiag


import numpy as np
from ....potential_fields.gravity import Simulation3DIntegral as Sim
from ....utils import sdiag, mkvc


def dask_getJtJdiag(self, m, W=None, f=None):
    """
    Return the diagonal of JtJ
    """

    self.model = m

    if W is None:
        W = np.ones(self.nD)
    else:
        W = W.diagonal()
    if getattr(self, "_gtg_diagonal", None) is None:
        diag = ((W[:, None] * self.Jmatrix) ** 2).sum(axis=0).compute()
        self._gtg_diagonal = diag
    else:
        diag = self._gtg_diagonal
    return mkvc((sdiag(np.sqrt(diag)) @ self.rhoDeriv).power(2).sum(axis=0))


Sim.getJtJdiag = dask_getJtJdiag



from discretize.utils import (  # noqa: F401
    rotation_matrix_from_normals,
    rotate_points_from_normals,
)
from .code_utils import deprecate_function

# deprecated functions
rotationMatrixFromNormals = deprecate_function(
    rotation_matrix_from_normals,
    "rotationMatrixFromNormals",
    removal_version="0.19.0",
    future_warn=True,
)
rotatePointsFromNormals = deprecate_function(
    rotate_points_from_normals,
    "rotatePointsFromNormals",
    removal_version="0.19.0",
    future_warn=True,
)



from ..survey import BaseSrc
from ..utils import Zero, validate_type
from ..base import BaseElectricalPDESimulation, BaseMagneticPDESimulation

__all__ = ["BaseEMSimulation", "BaseEMSrc"]


###############################################################################
#                                                                             #
#                             Base EM Simulation                                 #
#                                                                             #
###############################################################################


class BaseEMSimulation(BaseElectricalPDESimulation, BaseMagneticPDESimulation):
    """Base electromagnetic simulation class"""

    def __init__(self, mesh, storeInnerProduct=True, **kwargs):
        super().__init__(mesh=mesh, **kwargs)
        self.storeInnerProduct = storeInnerProduct

    @property
    def storeInnerProduct(self):
        """Whether to store inner product matrices

        Returns
        -------
        bool
        """
        return self._storeInnerProduct

    @storeInnerProduct.setter
    def storeInnerProduct(self, value):
        self._storeInnerProduct = validate_type("storeInnerProduct", value, bool)

    ####################################################
    # Make A Symmetric
    ####################################################
    @property
    def _makeASymmetric(self):
        if getattr(self, "__makeASymmetric", None) is None:
            self.__makeASymmetric = True
        return self.__makeASymmetric


###############################################################################
#                                                                             #
#                             Base EM Source                                  #
#                                                                             #
###############################################################################


class BaseEMSrc(BaseSrc):
    """Base class for a electromagnetic sources

    Parameters
    ----------
    location : (n_dim) numpy.ndarray
        Location of the source
    receiver_list : list of SimPEG.survey.BaseRx objects
        Sets the receivers associated with the source
    uid : uuid.UUID
        A universally unique identifier
    integrate : bool
        If ``True``, we integrate the source term
    """

    def __init__(self, receiver_list=None, location=None, integrate=False, **kwargs):
        super(BaseEMSrc, self).__init__(
            receiver_list=receiver_list, location=location, **kwargs
        )
        self.integrate = integrate

    @property
    def integrate(self):
        """Integrated source term

        Returns
        -------
        bool
            If ``True``, the source term is integrated
        """
        return self._integrate

    @integrate.setter
    def integrate(self, var):
        if not isinstance(var, bool):
            raise TypeError(f"integrate property is a bool. Got {type(var)}")
        self._integrate = var

    def eval(self, simulation):  # noqa: A003
        """Return magnetic and electric source terms

        Parameters
        ----------
        simulation : SimPEG.electromagnetics.base.BaseEMSimulation
            An instance of an electromagnetic simulation

        Returns
        -------
        tuple
            A tuple (s_m, s_e), where s_m is the discretized magnetic source term
            and s_e is the discretized electric course term.
        """
        s_m = self.s_m(simulation)
        s_e = self.s_e(simulation)
        return s_m, s_e

    def evalDeriv(self, simulation, v=None, adjoint=False):
        """Return derivative of the magnetic and electric source terms with respect to the model.

        Parameters
        ----------
        simulation : SimPEG.electromagnetics.base.BaseEMSimulation
            An instance of an electromagnetic simulation
        v : np.ndarray
            A vector
        adjoint : bool
            If ``True``, return the adjoint operation

        Returns
        -------
        tuple
            A tuple (s_mDeriv, s_eDerive). If `v` is not ``None``, the method returns
            the derivatives of the magnetic and electric sources times the vector `v`.
            If `v` is ``None``, the method returns the functions for multiplying the
            derivatives with a vector.
        """
        if v is not None:
            return (
                self.s_mDeriv(simulation, v, adjoint),
                self.s_eDeriv(simulation, v, adjoint),
            )
        else:
            return (
                lambda v: self.s_mDeriv(simulation, v, adjoint),
                lambda v: self.s_eDeriv(simulation, v, adjoint),
            )

    def s_m(self, simulation):
        """Magnetic source term

        Parameters
        ----------
        simulation : BaseEMSimulation
            An EM Simulation object

        Returns
        -------
        numpy.ndarray
            Magnetic source term defined on the mesh. Defined on faces for EB formulations,
            and defined on edges for HJ formulations.
        """
        return Zero()

    def s_e(self, simulation):
        """Electric source term

        Parameters
        ----------
        simulation : BaseEMSimulation
            An EM Simulation object

        Returns
        -------
        numpy.ndarray
            Electric source term defined on the mesh. Defined on edges for EB formulations,
            and defined on faces for HJ formulations.
        """
        return Zero()

    def s_mDeriv(self, simulation, v, adjoint=False):
        """
        Derivative of magnetic source term with respect to the inversion model

        Parameters
        ----------
        simulation : BaseEMSimulation
            An EM Simulation object
        v : numpy.ndarray
            A vector to take the dot product with
        adjoint : bool, default==Fasel
            If ``True``, return the adjoint operation

        Returns
        -------
        numpy.ndarray
            Product of the derivative of the magnetic source term and a vector
        """

        return Zero()

    def s_eDeriv(self, simulation, v, adjoint=False):
        """
        Derivative of electric source term with respect to the inversion model

        Parameters
        ----------
        simulation : BaseEMSimulation
            An EM Simulation object
        v : numpy.ndarray
            A vector to take the dot product with
        adjoint : bool, default==Fasel
            If ``True``, return the adjoint operation

        Returns
        -------
        numpy.ndarray
            Product of the derivative of the electric source term and a vector
        """
        return Zero()



import time
import numpy as np
from functools import wraps


class Counter(object):
    """Class for recording iterations and operation times.
    Counter allows anything that calls it to record iterations and
    timings in a simple way.

    Also has plotting functions that allow quick recalls of data.

    Examples
    --------
    If you want to use this, import *count* or *timeIt* and use them as
    decorators on class methods.


    >>> from SimPEG.utils import Counter, count, timeIt

    >>> class MyClass(object):
    ...     def __init__(self, url):
    ...         self.counter = Counter()
    ...     @count
    ...     def MyMethod(self):
    ...         pass
    ...     @timeIt
    ...     def MySecondMethod(self):
    ...         pass

    >>> c = MyClass('blah')
    >>> for i in range(100): c.MyMethod()
    >>> for i in range(300): c.MySecondMethod()
    >>> c.counter.summary()
    Counters:
      MyClass.MyMethod                        :      100
    Times:                                        mean      sum
      MyClass.MySecondMethod                  : 9.08e-07, 2.72e-04,  300x

    """

    def __init__(self):
        self._countList = {}
        self._timeList = {}

    def count(self, prop):
        """Increases the count of the property.

        Parameters
        ----------
        prop : str
            The property being counted
        """
        assert isinstance(prop, str), "The property must be a string."
        if prop not in self._countList:
            self._countList[prop] = 0
        self._countList[prop] += 1

    def countTic(self, prop):
        """Times a property call, this is the init call.

        Parameters
        ----------
        prop : str
            The property being timed
        """
        assert isinstance(prop, str), "The property must be a string."
        if prop not in self._timeList:
            self._timeList[prop] = []
        self._timeList[prop].append(-time.time())

    def countToc(self, prop):
        """Times a property call, this is the end call.

        Parameters
        ----------
        prop : str
            The property being timed
        """
        assert isinstance(prop, str), "The property must be a string."
        assert prop in self._timeList, "The property must already be in the dictionary."
        self._timeList[prop][-1] += time.time()

    def summary(self):
        """
        Provides a text summary of the current counters and timers.
        """
        print("Counters:")
        for prop in sorted(self._countList):
            print("  {0:<40}: {1:8d}".format(prop, self._countList[prop]))
        print("\nTimes:" + " " * 40 + "mean      sum")
        for prop in sorted(self._timeList):
            l = len(self._timeList[prop])
            a = np.array(self._timeList[prop])
            print(
                "  {0:<40}: {1:4.2e}, {2:4.2e}, {3:4d}x".format(
                    prop, a.mean(), a.sum(), l
                )
            )


def count(f):
    """Count function (**DOCSTRING INCOMPLETE**)

    Returns
    -------
    wrapper
        Returns the wrapper
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        counter = getattr(self, "counter", None)
        if isinstance(counter, Counter):
            counter.count(self.__class__.__name__ + "." + f.__name__)
        out = f(self, *args, **kwargs)
        return out

    return wrapper


def timeIt(f):
    """Timing function (**DOCSTRING INCOMPLETE**)

    Returns
    -------
    wrapper
        Returns the wrapper
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        counter = getattr(self, "counter", None)
        if isinstance(counter, Counter):
            counter.countTic(self.__class__.__name__ + "." + f.__name__)
        out = f(self, *args, **kwargs)
        if isinstance(counter, Counter):
            counter.countToc(self.__class__.__name__ + "." + f.__name__)
        return out

    return wrapper



import numpy as np
import scipy.sparse as sp
import gc
from .data_misfit import BaseDataMisfit
from .regularization import BaseRegularization, WeightedLeastSquares, Sparse
from .objective_function import BaseObjectiveFunction, ComboObjectiveFunction
from .optimization import Minimize
from .utils import (
    call_hooks,
    timeIt,
    Counter,
    validate_float,
    validate_type,
    validate_ndarray_with_shape,
)
from .simulation import DefaultSolver


class BaseInvProblem:
    """BaseInvProblem(dmisfit, reg, opt)"""

    def __init__(
        self, dmisfit, reg, opt, beta=1.0, debug=False, counter=None, **kwargs
    ):
        super().__init__(**kwargs)
        assert isinstance(reg, BaseRegularization) or isinstance(
            reg, BaseObjectiveFunction
        ), "reg must be a Regularization or Objective Function class."

        self.dmisfit = dmisfit

        self.reg = reg
        self.opt = opt
        self.beta = beta
        self.debug = debug
        self.counter = counter
        self.model = None
        # TODO: Remove: (and make iteration printers better!)
        self.opt.parent = self
        self.reg.parent = self
        self.dmisfit.parent = self

    #: Trade-off parameter
    @property
    def beta(self):
        """Trade-off parameter

        Returns
        -------
        float
        """
        return self._beta

    @beta.setter
    def beta(self, value):
        self._beta = validate_float("beta", value, min_val=0.0)

    @property
    def debug(self):
        """Whether to print debugging information.

        Returns
        -------
        bool
        """
        return self._debug

    @debug.setter
    def debug(self, value):
        self._debug = validate_type("debug", value, bool)

    @property
    def counter(self):
        """Set this to a `SimPEG.utils.Counter` if you want to count things.

        Returns
        -------
        None or SimPEG.utils.Counter
        """
        return self._counter

    @counter.setter
    def counter(self, value):
        if value is not None:
            value = validate_type("counter", value, Counter, cast=False)
        self._counter = value

    @property
    def dmisfit(self):
        """The data misfit.

        Returns
        -------
        SimPEG.objective_function.ComboObjectiveFunction
        """
        return self._dmisfit

    @dmisfit.setter
    def dmisfit(self, value):
        value = validate_type("dmisfit", value, BaseObjectiveFunction, cast=False)
        if not isinstance(value, ComboObjectiveFunction):
            value = ComboObjectiveFunction(objfcts=[value])
        self._dmisfit = value

    @property
    def reg(self):
        """The regularization object for the inversion

        Returns
        -------
        SimPEG.objective_function.ComboObjectiveFunction
        """
        return self._reg

    @reg.setter
    def reg(self, value):
        value = validate_type("reg", value, BaseObjectiveFunction, cast=False)
        if not isinstance(value, ComboObjectiveFunction):
            value = ComboObjectiveFunction(objfcts=[value])
        self._reg = value

    @property
    def opt(self):
        """The optimization routine.

        Returns
        -------
        SimPEG.optimization.Minimize
        """
        return self._opt

    @opt.setter
    def opt(self, value):
        self._opt = validate_type("opt", value, Minimize, cast=False)

    @property
    def deleteTheseOnModelUpdate(self):
        """A list of properties stored on this object to delete when the model is updated

        Returns
        -------
        list of str
            For example `['_MeSigma', '_MeSigmaI']`.
        """
        return []

    @property
    def model(self):
        """The inversion model.

        Returns
        -------
        numpy.ndarray
        """
        return self._model

    @model.setter
    def model(self, value):
        if value is not None:
            value = validate_ndarray_with_shape(
                "model", value, shape=[("*",), ("*", "*")], dtype=None
            )
        for prop in self.deleteTheseOnModelUpdate:
            if hasattr(self, prop):
                delattr(self, prop)
        self._model = value

    @call_hooks("startup")
    def startup(self, m0):
        """startup(m0)

        Called when inversion is first starting.
        """
        if self.debug:
            print("Calling InvProblem.startup")

        for fct in self.reg.objfcts:
            if (
                hasattr(fct, "reference_model")
                and getattr(fct, "reference_model", None) is None
            ):
                print(
                    "SimPEG.InvProblem will set Regularization.reference_model to m0."
                )
                fct.reference_model = m0

        self.phi_d = np.nan
        self.phi_m = np.nan

        self.model = m0

        solver = DefaultSolver
        set_default = True
        for objfct in self.dmisfit.objfcts:
            if (
                isinstance(objfct, BaseDataMisfit)
                and getattr(objfct.simulation, "solver", None) is not None
            ):
                solver = objfct.simulation.solver
                solver_opts = objfct.simulation.solver_opts
                print(
                    """
                        SimPEG.InvProblem is setting bfgsH0 to the inverse of the eval2Deriv.
                        ***Done using same Solver, and solver_opts as the {} problem***
                        """.format(
                        objfct.simulation.__class__.__name__
                    )
                )
                set_default = False
                break
        if set_default:
            print(
                """
                    SimPEG.InvProblem is setting bfgsH0 to the inverse of the eval2Deriv.
                    ***Done using the default solver {} and no solver_opts.***
                    """.format(
                    DefaultSolver.__name__
                )
            )
            solver = DefaultSolver
            solver_opts = {}

        self.opt.bfgsH0 = solver(
            sp.csr_matrix(self.reg.deriv2(self.model)), **solver_opts
        )

    @property
    def warmstart(self):
        return getattr(self, "_warmstart", [])

    @warmstart.setter
    def warmstart(self, value):
        assert type(value) is list, "warmstart must be a list."
        for v in value:
            assert type(v) is tuple, "warmstart must be a list of tuples (m, u)."
            assert (
                len(v) == 2
            ), "warmstart must be a list of tuples (m, u). YOURS IS NOT LENGTH 2!"
            assert isinstance(
                v[0], np.ndarray
            ), "first warmstart value must be a model."
        self._warmstart = value

    def getFields(self, m, store=False, deleteWarmstart=True):
        f = None

        for mtest, u_ofmtest in self.warmstart:
            if m is mtest:
                f = u_ofmtest
                if self.debug:
                    print("InvProb is Warm Starting!")
                break

        if f is None:
            if isinstance(self.dmisfit, BaseDataMisfit):
                f = self.dmisfit.simulation.fields(m)

            elif isinstance(self.dmisfit, BaseObjectiveFunction):
                f = []
                for objfct in self.dmisfit.objfcts:
                    if hasattr(objfct, "simulation"):
                        f += [objfct.simulation.fields(m)]
                    else:
                        f += []

        if deleteWarmstart:
            self.warmstart = []
        if store:
            self.warmstart += [(m, f)]

        return f

    def get_dpred(self, m, f):
        dpred = []
        for i, objfct in enumerate(self.dmisfit.objfcts):
            if hasattr(objfct, "simulation"):
                dpred += [objfct.simulation.dpred(m, f=f[i])]
            else:
                dpred += []
        return np.hstack(dpred)

    @timeIt
    def evalFunction(self, m, return_g=True, return_H=True):
        """evalFunction(m, return_g=True, return_H=True)"""

        self.model = m
        gc.collect()

        # Store fields if doing a line-search
        f = self.getFields(m, store=(return_g is False and return_H is False))

        # if isinstance(self.dmisfit, BaseDataMisfit):
        phi_d = self.dmisfit(m, f=f)
        self.dpred = self.get_dpred(m, f=f)

        phi_m = self.reg(m)

        self.phi_d, self.phi_d_last = phi_d, self.phi_d
        self.phi_m, self.phi_m_last = phi_m, self.phi_m

        # Only works for WeightedLeastSquares regularization
        if self.opt.print_type == "ubc":
            self.phi_s = 0.0
            self.phi_x = 0.0
            self.phi_y = 0.0
            self.phi_z = 0.0

            if not isinstance(self.reg, WeightedLeastSquares):
                regs = self.reg.objfcts
                mults = self.reg.multipliers
            else:
                regs = [self.reg]
                mults = [1.0]
            for reg, mult in zip(regs, mults):
                if isinstance(reg, Sparse):
                    i_s, i_x, i_y, i_z = 0, 1, 2, 3
                else:
                    i_s, i_x, i_y, i_z = 0, 1, 3, 5
                dim = reg.regularization_mesh.dim
                self.phi_s += mult * reg.objfcts[i_s](m) * reg.alpha_s
                self.phi_x += mult * reg.objfcts[i_x](m) * reg.alpha_x
                if dim > 1:
                    self.phi_z += mult * reg.objfcts[i_y](m) * reg.alpha_y
                if dim > 2:
                    self.phi_y = self.phi_z
                    self.phi_z += mult * reg.objfcts[i_z](m) * reg.alpha_z

        phi = phi_d + self.beta * phi_m

        out = (phi,)
        if return_g:
            phi_dDeriv = self.dmisfit.deriv(m, f=f)
            phi_mDeriv = self.reg.deriv(m)

            g = phi_dDeriv + self.beta * phi_mDeriv
            out += (g,)

        if return_H:

            def H_fun(v):
                phi_d2Deriv = self.dmisfit.deriv2(m, v, f=f)
                phi_m2Deriv = self.reg.deriv2(m, v=v)

                return phi_d2Deriv + self.beta * phi_m2Deriv

            H = sp.linalg.LinearOperator((m.size, m.size), H_fun, dtype=m.dtype)
            out += (H,)
        return out if len(out) > 1 else out[0]



import numpy as np
from scipy.optimize import minimize
import warnings


from ....utils import (
    mkvc,
    sdiag,
    Zero,
    validate_type,
    validate_string,
    validate_integer,
    validate_active_indices,
)
from ....base import BaseElectricalPDESimulation
from ....data import Data

from .survey import Survey
from .fields_2d import Fields2D, Fields2DCellCentered, Fields2DNodal
from .fields import FieldsDC, Fields3DCellCentered, Fields3DNodal
from .utils import _mini_pole_pole
from scipy.special import k0e, k1e, k0
from discretize.utils import make_boundary_bool


class BaseDCSimulation2D(BaseElectricalPDESimulation):
    """
    Base 2.5D DC problem
    """

    fieldsPair = Fields2D  # SimPEG.EM.Static.Fields_2D
    fieldsPair_fwd = FieldsDC
    # there's actually nT+1 fields, so we don't need to store the last one
    _mini_survey = None

    def __init__(
        self,
        mesh,
        survey=None,
        nky=11,
        storeJ=False,
        miniaturize=False,
        do_trap=False,
        fix_Jmatrix=False,
        surface_faces=None,
        **kwargs,
    ):
        super().__init__(mesh=mesh, survey=survey, **kwargs)
        self.nky = nky
        self.storeJ = storeJ
        self.fix_Jmatrix = fix_Jmatrix
        self.surface_faces = surface_faces

        do_trap = validate_type("do_trap", do_trap, bool)
        if not do_trap:
            # try to find an optimal set of quadrature points and weights
            def get_phi(r):
                e = np.ones_like(r)

                def phi(k):
                    # use log10 transform to enforce positivity
                    k = 10**k
                    A = r[:, None] * k0(r[:, None] * k)
                    v_i = A @ np.linalg.solve(A.T @ A, A.T @ e)
                    dv = (e - v_i) / len(r)
                    return np.linalg.norm(dv)

                def g(k):
                    A = r[:, None] * k0(r[:, None] * k)
                    return np.linalg.solve(A.T @ A, A.T @ e)

                return phi, g

            # find the minimum cell spacing, and the maximum side of the mesh
            min_r = min(self.mesh.edge_lengths)
            max_r = max(
                np.max(self.mesh.nodes, axis=0) - np.min(self.mesh.nodes, axis=0)
            )
            # generate test points log spaced between these two end members
            rs = np.logspace(np.log10(min_r / 4), np.log10(max_r * 4), 100)

            min_rinv = -np.log10(rs).max()
            max_rinv = -np.log10(rs).min()
            # a decent initial guess of the k_i's for the optimization = 1/rs
            k_i = np.linspace(min_rinv, max_rinv, self.nky)

            # these functions depend on r, so grab them
            func, g_func = get_phi(rs)

            # just use scipy's minimize for ease
            out = minimize(func, k_i)
            if self.verbose:
                print(f"optimized ks converged? : {out['success']}")
                print(f"Estimated transform Error: {out['fun']}")
            # transform the solution back to normal points
            points = 10 ** out["x"]
            # transform has a 2/pi and we want 1/pi, so divide by 2
            weights = g_func(points) / 2
            if not out["success"]:
                warnings.warn(
                    "Falling back to trapezoidal for integration. "
                    "You may need to change nky."
                )
                do_trap = True
        if do_trap:
            if self.verbose:
                print("doing trap")
            y = 0.0

            points = np.logspace(-4, 1, self.nky)
            dky = np.diff(points) / 2
            weights = np.r_[dky, 0] + np.r_[0, dky]
            weights *= np.cos(points * y)  # *(1.0/np.pi)
            # assume constant value at 0 frequency?
            weights[0] += points[0] / 2 * (1.0 + np.cos(points[0] * y))
            weights /= np.pi

        self._quad_weights = weights
        self._quad_points = points

        self.Ainv = [None for i in range(self.nky)]
        self.nT = self.nky - 1  # Only for using TimeFields

        # Do stuff to simplify the forward and JTvec operation if number of dipole
        # sources is greater than the number of unique pole sources
        miniaturize = validate_type("miniaturize", miniaturize, bool)
        if miniaturize:
            self._dipoles, self._invs, self._mini_survey = _mini_pole_pole(self.survey)

    @property
    def survey(self):
        """The DC survey object.

        Returns
        -------
        SimPEG.electromagnetics.static.resistivity.survey.Survey
        """
        if self._survey is None:
            raise AttributeError("Simulation must have a survey")
        return self._survey

    @survey.setter
    def survey(self, value):
        if value is not None:
            value = validate_type("survey", value, Survey, cast=False)
        self._survey = value

    @property
    def nky(self):
        """Number of kys to use in wavenumber space.

        Returns
        -------
        int
        """
        return self._nky

    @nky.setter
    def nky(self, value):
        self._nky = validate_integer("nky", value, min_val=3)

    @property
    def storeJ(self):
        """Whether to store the sensitivity matrix

        Returns
        -------
        bool
        """
        return self._storeJ

    @storeJ.setter
    def storeJ(self, value):
        self._storeJ = validate_type("storeJ", value, bool)

    @property
    def fix_Jmatrix(self):
        """Whether to fix the sensitivity matrix between iterations.

        Returns
        -------
        bool
        """
        return self._fix_Jmatrix

    @fix_Jmatrix.setter
    def fix_Jmatrix(self, value):
        self._fix_Jmatrix = validate_type("fix_Jmatrix", value, bool)

    @property
    def surface_faces(self):
        """Array defining which faces to interpret as surfaces of Neumann boundary

        DC problems will always enforce a Neumann boundary on surface interfaces.
        The default (available on semi-structured grids) assumes the top interface
        is the surface.

        Returns
        -------
        None or numpy.ndarray of bool
        """
        return self._surface_faces

    @surface_faces.setter
    def surface_faces(self, value):
        if value is not None:
            n_bf = self.mesh.boundary_faces.shape[0]
            value = validate_active_indices("surface_faces", value, n_bf)
        self._surface_faces = value

    def fields(self, m=None):
        if self.verbose:
            print(">> Compute fields")
        if m is not None:
            self.model = m
        if self.Ainv[0] is not None:
            for i in range(self.nky):
                self.Ainv[i].clean()
        f = self.fieldsPair(self)
        kys = self._quad_points
        f._quad_weights = self._quad_weights
        for iky, ky in enumerate(kys):
            A = self.getA(ky)
            if self.Ainv[iky] is not None:
                self.Ainv[iky].clean()
            self.Ainv[iky] = self.solver(A, **self.solver_opts)
            RHS = self.getRHS(ky)
            u = self.Ainv[iky] * RHS
            f[:, self._solutionType, iky] = u
        return f

    def fields_to_space(self, f, y=0.0):
        f_fwd = self.fieldsPair_fwd(self)
        phi = f[:, self._solutionType, :].dot(self._quad_weights)
        f_fwd[:, self._solutionType] = phi
        return f_fwd

    def dpred(self, m=None, f=None):
        """
        Project fields to receiver locations
        :param Fields u: fields object
        :rtype: numpy.ndarray
        :return: data
        """
        if f is None:
            if m is None:
                m = self.model
            f = self.fields(m)

        weights = self._quad_weights
        if self._mini_survey is not None:
            survey = self._mini_survey
        else:
            survey = self.survey

        temp = np.empty(survey.nD)
        count = 0
        for src in survey.source_list:
            for rx in src.receiver_list:
                d = rx.eval(src, self.mesh, f).dot(weights)
                temp[count : count + len(d)] = d
                count += len(d)

        return self._mini_survey_data(temp)

    def getJ(self, m, f=None):
        """
        Generate Full sensitivity matrix
        """
        if getattr(self, "_Jmatrix", None) is None:
            if self.verbose:
                print("Calculating J and storing")
            self.model = m
            if f is None:
                f = self.fields(m)
            self._Jmatrix = (self._Jtvec(m, v=None, f=f)).T
        return self._Jmatrix

    def Jvec(self, m, v, f=None):
        """
        Compute sensitivity matrix (J) and vector (v) product.
        """
        if self.storeJ:
            J = self.getJ(m, f=f)
            Jv = mkvc(np.dot(J, v))
            return Jv

        self.model = m

        if f is None:
            f = self.fields(m)

        if self._mini_survey is not None:
            survey = self._mini_survey
        else:
            survey = self.survey

        kys = self._quad_points
        weights = self._quad_weights

        Jv = np.zeros(survey.nD)
        # Assume y=0.
        # This needs some thoughts to implement in general when src is dipole

        # TODO: this loop is pretty slow .. (Parellize)
        for iky, ky in enumerate(kys):
            u_ky = f[:, self._solutionType, iky]
            count = 0
            for i_src, src in enumerate(survey.source_list):
                u_src = u_ky[:, i_src]
                dA_dm_v = self.getADeriv(ky, u_src, v, adjoint=False)
                # dRHS_dm_v = self.getRHSDeriv(ky, src, v) = 0
                du_dm_v = self.Ainv[iky] * (-dA_dm_v)  # + dRHS_dm_v)
                for rx in src.receiver_list:
                    df_dmFun = getattr(f, "_{0!s}Deriv".format(rx.projField), None)
                    df_dm_v = df_dmFun(iky, src, du_dm_v, v, adjoint=False)
                    Jv1_temp = rx.evalDeriv(src, self.mesh, f, df_dm_v)
                    # Trapezoidal intergration
                    Jv[count : count + len(Jv1_temp)] += weights[iky] * Jv1_temp
                    count += len(Jv1_temp)

        return self._mini_survey_data(Jv)

    def Jtvec(self, m, v, f=None):
        """
        Compute adjoint sensitivity matrix (J^T) and vector (v) product.
        """
        if self.storeJ:
            J = self.getJ(m, f=f)
            Jtv = mkvc(np.dot(J.T, v))
            return Jtv

        self.model = m

        if f is None:
            f = self.fields(m)

        return self._Jtvec(m, v=v, f=f)

    def _Jtvec(self, m, v=None, f=None):
        """
        Compute adjoint sensitivity matrix (J^T) and vector (v) product.
        Full J matrix can be computed by inputing v=None
        """
        kys = self._quad_points
        weights = self._quad_weights
        if self._mini_survey is not None:
            survey = self._mini_survey
        else:
            survey = self.survey

        if v is not None:
            # Ensure v is a data object.
            if isinstance(v, Data):
                v = v.dobs
            v = self._mini_survey_dataT(v)
            Jtv = np.zeros(m.size, dtype=float)

            for iky, ky in enumerate(kys):
                u_ky = f[:, self._solutionType, iky]
                count = 0
                for i_src, src in enumerate(survey.source_list):
                    u_src = u_ky[:, i_src]
                    df_duT_sum = 0
                    df_dmT_sum = 0
                    for rx in src.receiver_list:
                        my_v = v[count : count + rx.nD]
                        count += rx.nD
                        # wrt f, need possibility wrt m
                        PTv = rx.evalDeriv(src, self.mesh, f, my_v, adjoint=True)
                        df_duTFun = getattr(f, "_{0!s}Deriv".format(rx.projField), None)
                        df_duT, df_dmT = df_duTFun(iky, src, None, PTv, adjoint=True)
                        df_duT_sum += df_duT
                        df_dmT_sum += df_dmT

                    ATinvdf_duT = self.Ainv[iky] * df_duT_sum

                    dA_dmT = self.getADeriv(ky, u_src, ATinvdf_duT, adjoint=True)
                    # dRHS_dmT = self.getRHSDeriv(ky, src, ATinvdf_duT,
                    #                            adjoint=True)
                    du_dmT = -dA_dmT  # + dRHS_dmT=0
                    Jtv += weights[iky] * (df_dmT + du_dmT).astype(float)
            return mkvc(Jtv)

        else:
            # This is for forming full sensitivity matrix
            Jt = np.zeros((self.model.size, survey.nD), order="F")
            for iky, ky in enumerate(kys):
                u_ky = f[:, self._solutionType, iky]
                istrt = 0
                for i_src, src in enumerate(survey.source_list):
                    u_src = u_ky[:, i_src]
                    for rx in src.receiver_list:
                        # wrt f, need possibility wrt m
                        PT = rx.evalDeriv(src, self.mesh, f).toarray().T
                        ATinvdf_duT = self.Ainv[iky] * PT

                        dA_dmT = self.getADeriv(ky, u_src, ATinvdf_duT, adjoint=True)
                        Jtv = -weights[iky] * dA_dmT  # RHS=0
                        iend = istrt + rx.nD
                        if rx.nD == 1:
                            Jt[:, istrt] += Jtv
                        else:
                            Jt[:, istrt:iend] += Jtv
                        istrt += rx.nD
            return (self._mini_survey_data(Jt.T)).T

    def getSourceTerm(self, ky):
        """
        takes concept of source and turns it into a matrix
        """
        """
        Evaluates the sources, and puts them in matrix form
        :rtype: (numpy.ndarray, numpy.ndarray)
        :return: q (nC or nN, nSrc)
        """

        if self._mini_survey is not None:
            Srcs = self._mini_survey.source_list
        else:
            Srcs = self.survey.source_list

        if self._formulation == "EB":
            n = self.mesh.nN
            # return NotImplementedError

        elif self._formulation == "HJ":
            n = self.mesh.nC

        q = np.zeros((n, len(Srcs)), order="F")

        for i, src in enumerate(Srcs):
            q[:, i] = src.eval(self)
        return q

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = super().deleteTheseOnModelUpdate
        if self.fix_Jmatrix:
            return toDelete
        return toDelete + ["_Jmatrix"]

    def _mini_survey_data(self, d_mini):
        if self._mini_survey is not None:
            out = d_mini[self._invs[0]]  # AM
            out[self._dipoles[0]] -= d_mini[self._invs[1]]  # AN
            out[self._dipoles[1]] -= d_mini[self._invs[2]]  # BM
            out[self._dipoles[0] & self._dipoles[1]] += d_mini[self._invs[3]]  # BN
        else:
            out = d_mini
        return out

    def _mini_survey_dataT(self, v):
        if self._mini_survey is not None:
            out = np.zeros(self._mini_survey.nD)
            # Need to use ufunc.at because there could be repeated indices
            # That need to be properly handled.
            np.add.at(out, self._invs[0], v)  # AM
            np.subtract.at(out, self._invs[1], v[self._dipoles[0]])  # AN
            np.subtract.at(out, self._invs[2], v[self._dipoles[1]])  # BM
            np.add.at(out, self._invs[3], v[self._dipoles[0] & self._dipoles[1]])  # BN
            return out
        else:
            out = v
        return out


class Simulation2DCellCentered(BaseDCSimulation2D):
    """
    2.5D cell centered DC problem
    """

    _solutionType = "phiSolution"
    _formulation = "HJ"  # CC potentials means J is on faces
    fieldsPair = Fields2DCellCentered
    fieldsPair_fwd = Fields3DCellCentered

    def __init__(self, mesh, survey=None, bc_type="Robin", **kwargs):
        super().__init__(mesh, survey=survey, **kwargs)
        V = sdiag(self.mesh.cell_volumes)
        self.Div = V @ self.mesh.face_divergence
        self.Grad = self.Div.T
        self.bc_type = bc_type

    @property
    def bc_type(self):
        """Type of boundary condition to use for simulation.

        Returns
        -------
        {"Dirichlet", "Neumann", "Robin"}
        """
        return self._bc_type

    @bc_type.setter
    def bc_type(self, value):
        self._bc_type = validate_string(
            "bc_type", value, ["Dirichlet", "Neumann", ("Robin", "Mixed")]
        )

    def getA(self, ky):
        """
        Make the A matrix for the cell centered DC resistivity problem
        A = D MfRhoI G
        """
        # To handle Mixed boundary condition
        self.setBC(ky=ky)
        D = self.Div
        G = self.Grad
        if self.bc_type != "Dirichlet":
            G = G - self._MBC[ky]
        MfRhoI = self.MfRhoI
        # Get resistivity rho
        A = D * MfRhoI * G + ky**2 * self.MccSigma
        if self.bc_type == "Neumann":
            A[0, 0] = A[0, 0] + 1.0
        return A

    def getADeriv(self, ky, u, v, adjoint=False):
        D = self.Div
        G = self.Grad
        if self.bc_type != "Dirichlet":
            G = G - self._MBC[ky]
        if adjoint:
            return self.MfRhoIDeriv(
                G * u, D.T * v, adjoint=adjoint
            ) + ky**2 * self.MccSigmaDeriv(u, v, adjoint=adjoint)
        else:
            return D * self.MfRhoIDeriv(
                G * u, v, adjoint=adjoint
            ) + ky**2 * self.MccSigmaDeriv(u, v, adjoint=adjoint)

    def getRHS(self, ky):
        """
        RHS for the DC problem
        q
        """

        RHS = self.getSourceTerm(ky)
        return RHS

    def getRHSDeriv(self, ky, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, ky, adjoint=adjoint)
        # return qDeriv
        return Zero()

    def setBC(self, ky=None):
        if self.bc_type == "Dirichlet":
            return
        if getattr(self, "_MBC", None) is None:
            self._MBC = {}
        if ky in self._MBC:
            # I have already created the BC matrix for this wavenumber
            return
        if self.bc_type == "Neumann":
            alpha, beta, gamma = 0, 1, 0
        else:
            mesh = self.mesh
            boundary_faces = mesh.boundary_faces
            boundary_normals = mesh.boundary_face_outward_normals
            n_bf = len(boundary_faces)

            # Top gets 0 Neumann
            alpha = np.zeros(n_bf)
            beta = np.ones(n_bf)
            gamma = 0

            # assume a source point at the middle of the top of the mesh
            middle = np.median(mesh.nodes, axis=0)
            top_v = np.max(mesh.nodes[:, -1])
            source_point = np.r_[middle[:-1], top_v]

            r_vec = boundary_faces - source_point
            r = np.linalg.norm(r_vec, axis=-1)
            r_hat = r_vec / r[:, None]
            r_dot_n = np.einsum("ij,ij->i", r_hat, boundary_normals)

            if self.surface_faces is None:
                # determine faces that are on the sides and bottom of the mesh...
                if mesh._meshType.lower() == "tree":
                    not_top = boundary_faces[:, -1] != top_v
                elif mesh._meshType.lower() in ["tensor", "curv"]:
                    # mesh faces are ordered, faces_x, faces_y, faces_z so...
                    is_b = make_boundary_bool(mesh.shape_faces_y)
                    is_t = np.zeros(mesh.shape_faces_y, dtype=bool, order="F")
                    is_t[:, -1] = True
                    is_t = is_t.reshape(-1, order="F")[is_b]
                    not_top = np.ones(boundary_faces.shape[0], dtype=bool)
                    not_top[-len(is_t) :] = ~is_t
                else:
                    raise NotImplementedError(
                        f"Unable to infer surface boundaries for {type(mesh)}, please "
                        f"set the `surface_faces` property."
                    )
            else:
                not_top = ~self.surface_faces

            # use the exponentialy scaled modified bessel function of second kind,
            # (the division will cancel out the scaling)
            # This is more stable for large values of ky * r
            # actual ratio is k1/k0...
            alpha[not_top] = (ky * k1e(ky * r) / k0e(ky * r) * r_dot_n)[not_top]

        B, bc = self.mesh.cell_gradient_weak_form_robin(alpha, beta, gamma)
        # bc should always be 0 because gamma was always 0 above
        self._MBC[ky] = B


class Simulation2DNodal(BaseDCSimulation2D):
    """
    2.5D nodal DC problem
    """

    _solutionType = "phiSolution"
    _formulation = "EB"  # CC potentials means J is on faces
    fieldsPair = Fields2DNodal
    fieldsPair_fwd = Fields3DNodal
    _gradT = None

    def __init__(self, mesh, survey=None, bc_type="Robin", **kwargs):
        super().__init__(mesh=mesh, survey=survey, **kwargs)
        self.solver_opts["is_symmetric"] = True
        self.solver_opts["is_positive_definite"] = True
        self.bc_type = bc_type

    @property
    def bc_type(self):
        """Type of boundary condition to use for simulation.

        Returns
        -------
        {"Neumann", "Robin"}

        Notes
        -----
        Robin and Mixed are equivalent.
        """
        return self._bc_type

    @bc_type.setter
    def bc_type(self, value):
        self._bc_type = validate_string(
            "bc_type", value, ["Neumann", ("Robin", "Mixed")]
        )

    def getA(self, ky):
        """
        Make the A matrix for the cell centered DC resistivity problem
        A = D MfRhoI G
        """
        # To handle Mixed boundary condition
        self.setBC(ky=ky)

        MeSigma = self.MeSigma
        MnSigma = self.MnSigma
        Grad = self.mesh.nodal_gradient
        if self._gradT is None:
            self._gradT = Grad.T.tocsr()  # cache the .tocsr()
        GradT = self._gradT
        A = GradT * MeSigma * Grad + ky**2 * MnSigma

        if self.bc_type != "Neumann":
            try:
                A = A + sdiag(self._AvgBC[ky] @ self.sigma)
            except ValueError as err:
                if len(self.sigma) != len(self.mesh):
                    raise NotImplementedError(
                        "Anisotropic conductivity is not supported for Robin boundary "
                        "conditions, please use 'Neumann'."
                    )
                else:
                    raise err
        return A

    def getADeriv(self, ky, u, v, adjoint=False):
        Grad = self.mesh.nodal_gradient

        if adjoint:
            out = self.MeSigmaDeriv(
                Grad * u, Grad * v, adjoint=adjoint
            ) + ky**2 * self.MnSigmaDeriv(u, v, adjoint=adjoint)
        else:
            out = Grad.T * self.MeSigmaDeriv(
                Grad * u, v, adjoint=adjoint
            ) + ky**2 * self.MnSigmaDeriv(u, v, adjoint=adjoint)
        if self.bc_type != "Neumann" and self.sigmaMap is not None:
            if getattr(self, "_MBC_sigma", None) is None:
                self._MBC_sigma = {}
            if ky not in self._MBC_sigma:
                self._MBC_sigma[ky] = self._AvgBC[ky] @ self.sigmaDeriv
            if not isinstance(u, Zero):
                u = u.flatten()
                if v.ndim > 1:
                    u = u[:, None]
                if not adjoint:
                    out += u * (self._MBC_sigma[ky] @ v)
                else:
                    out += self._MBC_sigma[ky].T @ (u * v)
        return out

    def getRHS(self, ky):
        """
        RHS for the DC problem
        q
        """

        RHS = self.getSourceTerm(ky)
        return RHS

    def getRHSDeriv(self, ky, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, ky, adjoint=adjoint)
        # return qDeriv
        return Zero()

    def setBC(self, ky=None):
        if self.bc_type == "Dirichlet":
            # do nothing
            raise ValueError(
                "Dirichlet conditions are not supported in the Nodal formulation"
            )
        elif self.bc_type == "Neumann":
            if self.verbose:
                print(
                    "Homogeneous Neumann is the natural BC for this Nodal discretization."
                )
            return
        else:
            if getattr(self, "_AvgBC", None) is None:
                self._AvgBC = {}
            if ky in self._AvgBC:
                return
            mesh = self.mesh
            # calculate alpha, beta, gamma at the boundary faces
            boundary_faces = mesh.boundary_faces
            boundary_normals = mesh.boundary_face_outward_normals
            n_bf = len(boundary_faces)

            alpha = np.zeros(n_bf)

            # assume a source point at the middle of the top of the mesh
            middle = np.median(mesh.nodes, axis=0)
            top_v = np.max(mesh.nodes[:, -1])
            source_point = np.r_[middle[:-1], top_v]

            r_vec = boundary_faces - source_point
            r = np.linalg.norm(r_vec, axis=-1) + 1e-16
            r_hat = r_vec / r[:, None]
            r_dot_n = np.einsum("ij,ij->i", r_hat, boundary_normals)

            if self.surface_faces is None:
                # determine faces that are on the sides and bottom of the mesh...
                if mesh._meshType.lower() == "tree":
                    not_top = boundary_faces[:, -1] != top_v
                elif mesh._meshType.lower() in ["tensor", "curv"]:
                    # mesh faces are ordered, faces_x, faces_y, faces_z so...
                    is_b = make_boundary_bool(mesh.shape_faces_y)
                    is_t = np.zeros(mesh.shape_faces_y, dtype=bool, order="F")
                    is_t[:, -1] = True
                    is_t = is_t.reshape(-1, order="F")[is_b]
                    not_top = np.ones(boundary_faces.shape[0], dtype=bool)
                    not_top[-len(is_t) :] = ~is_t
                else:
                    raise NotImplementedError(
                        f"Unable to infer surface boundaries for {type(mesh)}, please "
                        f"set the `surface_faces` property."
                    )
            else:
                not_top = ~self.surface_faces

            # use the exponentiall scaled modified bessel function of second kind,
            # (the division will cancel out the scaling)
            # This is more stable for large values of ky * r
            # actual ratio is k1/k0...
            alpha[not_top] = (ky * k1e(ky * r) / k0e(ky * r) * r_dot_n)[not_top]

            P_bf = self.mesh.project_face_to_boundary_face

            AvgN2Fb = P_bf @ self.mesh.average_node_to_face
            AvgCC2Fb = P_bf @ self.mesh.average_cell_to_face

            AvgCC2Fb = sdiag(alpha * (P_bf @ self.mesh.face_areas)) @ AvgCC2Fb
            self._AvgBC[ky] = AvgN2Fb.T @ AvgCC2Fb


Simulation2DCellCentred = Simulation2DCellCentered  # UK and US



import numpy as np
import scipy.ndimage as ndi
import scipy.sparse as sp
from .mat_utils import mkvc
from scipy.spatial import Delaunay
from .code_utils import deprecate_function
from discretize.base import BaseMesh


def add_block(cell_centers, model, p0, p1, prop_value):
    """Add a homogeneous block to an existing cell centered model

    Parameters
    ----------
    cell_centers : (n_cells, dim) numpy.ndarray or discretize.base.BaseMesh
        A mesh or its gridded cell center locations
    model : (n_cells) numpy.ndarray
        Cell-centered model. Currently, this is only implemented for isotropic properties.
    p0 : (dim) numpy.ndarray
        Bottom southwest corner of the block
    p1 : (dim) numpy.ndarray
        Top northeast corner of the block
    prop_value : float
        Physical property value assigned to the block

    Returns
    -------
    (n_cells) numpy.ndarray
        The updated cell-centered model which includes the block
    """
    if isinstance(cell_centers, BaseMesh):
        cell_centers = cell_centers.cell_centers

    ind = get_indices_block(p0, p1, cell_centers)
    new_model = model.copy()
    new_model[ind] = prop_value
    return new_model


def get_indices_block(p0, p1, cell_centers):
    """Get indices for cells whose centers lie inside specified block

    Parameters
    ----------
    p0 : (dim) numpy.ndarray
        Bottom southwest corner of the block
    p1 : (dim) numpy.ndarray
        Top northeast corner of the block
    cell_centers : (n_cells, dim) numpy.ndarray or discretize.base.BaseMesh
        A mesh or its gridded cell center locations

    Returns
    -------
    tuple of int
        Indices of the cells whose center lie within the specified block

    """

    if isinstance(cell_centers, BaseMesh):
        cell_centers = cell_centers.cell_centers

    # Validation: p0 and p1 live in the same dimensional space
    assert len(p0) == len(p1), "Dimension mismatch. len(p0) != len(p1)"

    # Validation: mesh and points live in the same dimensional space
    dimMesh = np.size(cell_centers[0, :])
    assert len(p0) == dimMesh, "Dimension mismatch. len(p0) != dimMesh"

    for ii in range(len(p0)):
        p0[ii], p1[ii] = np.min([p0[ii], p1[ii]]), np.max([p0[ii], p1[ii]])

    if dimMesh == 1:
        # Define the reference points
        x1 = p0[0]
        x2 = p1[0]

        indX = (x1 <= cell_centers[:, 0]) & (cell_centers[:, 0] <= x2)
        ind = np.where(indX)

    elif dimMesh == 2:
        # Define the reference points
        x1 = p0[0]
        y1 = p0[1]

        x2 = p1[0]
        y2 = p1[1]

        indX = (x1 <= cell_centers[:, 0]) & (cell_centers[:, 0] <= x2)
        indY = (y1 <= cell_centers[:, 1]) & (cell_centers[:, 1] <= y2)

        ind = np.where(indX & indY)

    elif dimMesh == 3:
        # Define the points
        x1 = p0[0]
        y1 = p0[1]
        z1 = p0[2]

        x2 = p1[0]
        y2 = p1[1]
        z2 = p1[2]

        indX = (x1 <= cell_centers[:, 0]) & (cell_centers[:, 0] <= x2)
        indY = (y1 <= cell_centers[:, 1]) & (cell_centers[:, 1] <= y2)
        indZ = (z1 <= cell_centers[:, 2]) & (cell_centers[:, 2] <= z2)

        ind = np.where(indX & indY & indZ)

    # Return a tuple
    return ind


def create_block_in_wholespace(
    cell_centers, p0, p1, background_value=0.0, block_value=1.0
):
    """Construct cell-centered model comprised of a block in a wholespace.

    Parameters
    ----------
    cell_centers : (n_cells, dim) numpy.ndarray or discretize.base.BaseMesh
        A mesh or its gridded cell center locations
    p0 : (dim) numpy.ndarray
        Bottom southwest corner of the block
    p1 : (dim) numpy.ndarray
        Top northeast corner of the block
    background_value : float, optional
        Background physical property value.
    block_value : float, optional
        Block physical property value

    Returns
    -------
    (n_cells) numpy.ndarray
        Physical property model defined at the cell centers
    """

    if isinstance(cell_centers, BaseMesh):
        cell_centers = cell_centers.cell_centers

    # Used to use a single input 'vals' for background and block
    try:
        background_value, block_value = background_value
    except TypeError:
        pass

    sigma = np.zeros(cell_centers.shape[0]) + background_value
    ind = getIndicesBlock(p0, p1, cell_centers)

    sigma[ind] = block_value

    return mkvc(sigma)


def create_ellipse_in_wholespace(
    cell_centers, center=None, anisotropy=None, slope=10.0, theta=0.0
):
    """Construct cell-centered model comprised of an ellipsoid in a wholespace.

    Parameters
    ----------
    cell_centers : (n_cells, dim) numpy.ndarray or discretize.base.BaseMesh
        A mesh or its gridded cell center locations
    center : (dim) numpy.ndarray, optional
        Center of the ellipsoid, default is [0, 0, 0]
    anisotropy : (dim) numpy.ndarray, optional
        Anisotropy, default is an isotropic elipse ([1, 1, 1]).
    slope : float, optional
        Slope
    theta : float, optional
        Angle

    Returns
    -------
    (n_cells) numpy.ndarray
        Physical property model defined at the cell centers
    """

    if isinstance(cell_centers, BaseMesh):
        cell_centers = cell_centers.cell_centers

    if center is None:
        center = [0, 0, 0]
    if anisotropy is None:
        anisotropy = [1, 1, 1]
    G = cell_centers.copy()
    dim = cell_centers.shape[1]
    for i in range(dim):
        G[:, i] = G[:, i] - center[i]

    theta = -theta * np.pi / 180
    M = np.array(
        [
            [np.cos(theta), -np.sin(theta), 0],
            [np.sin(theta), np.cos(theta), 0],
            [0, 0, 1.0],
        ]
    )
    M = M[:dim, :dim]
    G = M.dot(G.T).T

    for i in range(dim):
        G[:, i] = G[:, i] / anisotropy[i] * 2.0

    D = np.sqrt(np.sum(G**2, axis=1))
    return -np.arctan((D - 1) * slope) * (2.0 / np.pi) / 2.0 + 0.5


def get_indices_sphere(center, radius, cell_centers):
    """Get indices for cells whose centers lie inside a sphere

    Parameters
    ----------
    center : (dim) numpy.ndarray
        Location of the center of the sphere
    radius : float
        Radius of the sphere
    cell_centers : (n_cells, dim) numpy.ndarray or discretize.base.BaseMesh
        A mesh or its gridded cell center locations

    Returns
    -------
    tuple of int
        Indices of the cells whose center lie within the specified sphere

    """

    if isinstance(cell_centers, BaseMesh):
        cell_centers = cell_centers.cell_centers

    # Validation: mesh and point (p0) live in the same dimensional space
    dimMesh = np.size(cell_centers[0, :])
    assert len(center) == dimMesh, "Dimension mismatch. len(p0) != dimMesh"

    if dimMesh == 1:
        # Define the reference points

        ind = np.abs(center[0] - cell_centers[:, 0]) < radius

    elif dimMesh == 2:
        # Define the reference points

        ind = (
            np.sqrt(
                (center[0] - cell_centers[:, 0]) ** 2
                + (center[1] - cell_centers[:, 1]) ** 2
            )
            < radius
        )

    elif dimMesh == 3:
        # Define the points
        ind = (
            np.sqrt(
                (center[0] - cell_centers[:, 0]) ** 2
                + (center[1] - cell_centers[:, 1]) ** 2
                + (center[2] - cell_centers[:, 2]) ** 2
            )
            < radius
        )

    # Return a tuple
    return ind


def create_2_layer_model(cell_centers, depth, top_value=1.0, bottom_value=0.0):
    """Create a basic two layered model

    This function creates a physical property model consisting of 2 layers.

    Parameters
    ----------
    cell_centers : (n_cells, dim) numpy.ndarray or discretize.base.BaseMesh
        A mesh or its gridded cell center locations
    depth : float
        Depth defining the interface between layer 1 and layer 2
    top_value : float, optional
        Physical property value for the top layer
    bottom_value : float, optional
        Physical property value for the bottom layer

    Returns
    -------
    (n_cells) numpy.ndarray
        Cell-centered physical property model

    """

    if isinstance(cell_centers, BaseMesh):
        cell_centers = cell_centers.cell_centers

    try:
        top_value, bottom_value = top_value
    except TypeError:
        pass

    sigma = np.zeros(cell_centers.shape[0]) + bottom_value

    dim = np.size(cell_centers[0, :])

    p0 = np.zeros(dim)
    p1 = np.zeros(dim)

    # Identify 1st cell centered reference point
    p0[0] = cell_centers[0, 0]
    if dim > 1:
        p0[1] = cell_centers[0, 1]
    if dim > 2:
        p0[2] = cell_centers[0, 2]

    # Identify the last cell-centered reference point
    p1[0] = cell_centers[-1, 0]
    if dim > 1:
        p1[1] = cell_centers[-1, 1]
    if dim > 2:
        p1[2] = cell_centers[-1, 2]

    # The depth is always defined on the last one.
    p1[len(p1) - 1] -= depth

    ind = getIndicesBlock(p0, p1, cell_centers)

    sigma[ind] = top_value

    return mkvc(sigma)


def create_from_function(cell_centers, fun_handle):
    """Define physical property model from scalar analytic function.

    For a function handle that defines a scalar physical property as function of
    location, **create_from_function** outputs the discrete representation
    of the physical property distribution on the mesh provided.

    Parameters
    ----------
    cell_centers : (n_cells, dim) numpy.ndarray or discretize.base.BaseMesh
        A mesh or its gridded cell center locations.
    fun_handle : function
        A function which defines a scalar physical property value as a function
        of location. The input argument for the location must be an (*n*, *dim*)
        ``numpy.ndarray``.

    Returns
    -------
    (n_cells) numpy.ndarray
        Cell-centered physical property model for the mesh.
    """

    if isinstance(cell_centers, BaseMesh):
        cell_centers = cell_centers.cell_centers

    dim = np.size(cell_centers[0, :])
    CC = [cell_centers[:, 0]]
    if dim > 1:
        CC.append(cell_centers[:, 1])
    if dim > 2:
        CC.append(cell_centers[:, 2])

    sigma = fun_handle(*CC)

    return mkvc(sigma)


def create_layers_model(cell_centers, layer_tops, layer_values):
    """Create physical property model consisting of a set of infinite horizontal layers.

    Parameters
    ----------
    cell_centers : (n_cells, dim) numpy.ndarray or discretize.base.BaseMesh
        A mesh or its gridded cell center locations
    layer_tops : (n_layer) numpy.ndarray
        Elevation values (z +ve up) for the top of each layer. Layers are defined from top to bottom.
        The first value can be very large if the top layer (e.g. air) extends to infinity.
    layer_values : (n_layer) numpy.ndarray
        Physical property value for each layer from top to bottom.

    Returns
    -------
    (n_cells) numpy.ndarray
        Cell-centered physical property model for the mesh.
    """

    if isinstance(cell_centers, BaseMesh):
        cell_centers = cell_centers.cell_centers

    descending = np.linalg.norm(sorted(layer_tops, reverse=True) - layer_tops) < 1e-20

    # TODO: put an error check to make sure that there is an ordering... needs to work with inf elts
    # assert ascending or descending, "Layers must be listed in either ascending or descending order"

    # start from bottom up
    if not descending:
        zprop = np.hstack([mkvc(layer_tops, 2), mkvc(layer_values, 2)])
        zprop.sort(axis=0)
        layer_tops, layer_values = zprop[::-1, 0], zprop[::-1, 1]

    # put in vector form
    layer_tops, layer_values = mkvc(layer_tops), mkvc(layer_values)

    # initialize with bottom layer
    dim = cell_centers.shape[1]
    if dim == 3:
        z = cell_centers[:, 2]
    elif dim == 2:
        z = cell_centers[:, 1]
    elif dim == 1:
        z = cell_centers[:, 0]

    model = np.zeros(cell_centers.shape[0])

    for i, top in enumerate(layer_tops):
        zind = z <= top
        model[zind] = layer_values[i]

    return model


def create_random_model(shape, seed=1000, anisotropy=None, its=100, bounds=None):
    """Create random model by convolving a kernel with a uniformly distributed random model.

    Parameters
    ----------
    shape : int or tuple of int
        Shape of the model. Can define a vector of size (n_cells) or define the dimensions of a tensor
    seed : int, optional
        If not None, sets the seed for the random uniform model that is convolved with the kernel.
    anisotropy : numpy.ndarray
        this is the (*3*, *n*) blurring kernel that is used.
    its : int
        Number of smoothing iterations after convolutions
    bounds : list of float
        Lower and upper bound for the model values

    Returns
    -------
    numpy.ndarray
        Physical property model


    Examples
    --------

    >>> import matplotlib.pyplot as plt
    >>> from SimPEG.utils.model_builder import create_random_model
    >>> m = create_random_model((50,50), bounds=[-4,0])
    >>> plt.colorbar(plt.imshow(m))
    >>> plt.title('A very cool, yet completely random model.')
    >>> plt.show()

    """
    if bounds is None:
        bounds = [0, 1]

    if seed is not None:
        np.random.seed(seed)
        print("Using a seed of: ", seed)

    if isinstance(shape, (int, float)):
        shape = (shape,)  # make it a tuple for consistency

    mr = np.random.rand(*shape)
    if anisotropy is None:
        if len(shape) == 1:
            smth = np.array([1, 10.0, 1], dtype=float)
        elif len(shape) == 2:
            smth = np.array([[1, 7, 1], [2, 10, 2], [1, 7, 1]], dtype=float)
        elif len(shape) == 3:
            kernal = np.array([1, 4, 1], dtype=float).reshape((1, 3))
            smth = np.array(
                sp.kron(sp.kron(kernal, kernal.T).todense()[:], kernal).todense()
            ).reshape((3, 3, 3))
    else:
        assert len(anisotropy.shape) is len(shape), "Anisotropy must be the same shape."
        smth = np.array(anisotropy, dtype=float)

    smth = smth / smth.sum()  # normalize
    mi = mr
    for _ in range(its):
        mi = ndi.convolve(mi, smth)

    # scale the model to live between the bounds.
    mi = (mi - mi.min()) / (mi.max() - mi.min())  # scaled between 0 and 1
    mi = mi * (bounds[1] - bounds[0]) + bounds[0]

    return mi


def get_indices_polygon(mesh, pts):
    """Get indices for cells whose centers lie within the convex hull of a set of points.

    Parameters
    ----------
    mesh : discretize.base.BaseMesh
        Discretize mesh
    pts : (n, dim) numpy.ndarray
        Set of points defining the convex hull

    Returns
    -------
    tuple of int
        Indices of the cells whose center lies within convex hull

    """
    if mesh.dim == 1:
        assert "Only works for a mesh greater than 1-dimension"
    elif mesh.dim == 2:
        assert ~(pts.shape[1] != 2), "Please input (*,2) array"
    elif mesh.dim == 3:
        assert ~(pts.shape[1] != 3), "Please input (*,3) array"
    hull = Delaunay(pts)
    inds = hull.find_simplex(mesh.cell_centers) >= 0
    return inds


################################################
#             DEPRECATED FUNCTIONS
################################################


addBlock = deprecate_function(add_block, "addBlock", removal_version="0.19.0")

getIndicesBlock = deprecate_function(
    get_indices_block, "getIndicesBlock", removal_version="0.19.0"
)

defineBlock = deprecate_function(
    create_block_in_wholespace, "defineBlock", removal_version="0.19.0"
)

defineEllipse = deprecate_function(
    create_ellipse_in_wholespace, "defineEllipse", removal_version="0.19.0"
)

getIndicesSphere = deprecate_function(
    get_indices_sphere, "getIndicesSphere", removal_version="0.19.0"
)

defineTwoLayers = deprecate_function(
    create_2_layer_model, "defineTwoLayers", removal_version="0.19.0"
)

layeredModel = deprecate_function(
    create_layers_model, "layeredModel", removal_version="0.19.0"
)

randomModel = deprecate_function(
    create_random_model, "randomModel", removal_version="0.19.0"
)

polygonInd = deprecate_function(
    get_indices_polygon, "polygonInd", removal_version="0.19.0"
)

scalarConductivity = deprecate_function(
    create_from_function, "scalarConductivity", removal_version="0.19.0"
)



import numpy as np
from scipy.constants import mu_0, pi
from scipy.special import erf
from SimPEG import utils


def hzAnalyticDipoleT(r, t, sigma):
    theta = np.sqrt((sigma * mu_0) / (4 * t))
    tr = theta * r
    etr = erf(tr)
    t1 = (9 / (2 * tr**2) - 1) * etr
    t2 = (1 / np.sqrt(pi)) * (9 / tr + 4 * tr) * np.exp(-(tr**2))
    hz = (t1 - t2) / (4 * pi * r**3)
    return hz


def hzAnalyticCentLoopT(a, t, sigma):
    theta = np.sqrt((sigma * mu_0) / (4 * t))
    ta = theta * a
    eta = erf(ta)
    t1 = (3 / (np.sqrt(pi) * ta)) * np.exp(-(ta**2))
    t2 = (1 - (3 / (2 * ta**2))) * eta
    hz = (t1 + t2) / (2 * a)
    return hz


def TransientMagneticDipoleWholeSpace(
    XYZ, srcLoc, sig, t, moment, fieldType="h", mu_r=1
):
    """
    Analytical solution for a dipole in a whole-space.

    """

    mu = 4 * np.pi * 1e-7 * mu_r

    if isinstance(moment, str):
        if moment == "X":
            mx, my, mz = 1.0, 0.0, 0.0
        elif moment == "Y":
            mx, my, mz = 0.0, 1.0, 0.0
        elif moment == "Z":
            mx, my, mz = 0.0, 0.0, 1.0
        else:
            raise NotImplementedError("String type for moment not recognized")

    else:
        mx, my, mz = moment[0], moment[1], moment[2]

    XYZ = utils.as_array_n_by_dim(XYZ, 3)

    dx = XYZ[:, 0] - srcLoc[0]
    dy = XYZ[:, 1] - srcLoc[1]
    dz = XYZ[:, 2] - srcLoc[2]

    r = np.sqrt(dx**2.0 + dy**2.0 + dz**2.0)
    k = np.sqrt(mu * sig / (4 * t))
    kr = k * r

    if fieldType == "h":
        front = 1 / (4.0 * pi * r**3.0)
        mid = 3 * erf(kr) - (4 * kr**3 + 6 * kr) * np.exp(-(kr**2)) / np.sqrt(pi)
        end = -erf(kr) + (4 * kr**3 + 2 * kr) * np.exp(-(kr**2)) / np.sqrt(pi)

        Fx = front * (
            mx * ((dx / r) ** 2.0 * mid + end)
            + my * ((dy * dx / r**2.0) * mid)
            + mz * ((dx * dz / r**2.0) * mid)
        )

        Fy = front * (
            mx * ((dx * dy / r**2.0) * mid)
            + my * ((dy / r) ** 2.0 * mid + end)
            + mz * ((dy * dz / r**2.0) * mid)
        )

        Fz = front * (
            mx * ((dx * dz / r**2.0) * mid)
            + my * ((dy * dz / r**2.0) * mid)
            + mz * ((dz / r) ** 2.0 * mid + end)
        )

    elif fieldType == "dhdt":
        front = (4 * k**5 / (pi**1.5 * mu * sig)) * np.exp(-(kr**2))
        mid = kr**2
        end = 1 - kr**2

        Fx = front * (
            mx * ((dx / r) ** 2.0 * mid + end)
            + my * ((dy * dx / r**2.0) * mid)
            + mz * ((dx * dz / r**2.0) * mid)
        )

        Fy = front * (
            mx * ((dx * dy / r**2.0) * mid)
            + my * ((dy / r) ** 2.0 * mid + end)
            + mz * ((dy * dz / r**2.0) * mid)
        )

        Fz = front * (
            mx * ((dx * dz / r**2.0) * mid)
            + my * ((dy * dz / r**2.0) * mid)
            + mz * ((dz / r) ** 2.0 * mid + end)
        )

    elif fieldType == "e":
        front = (2 * k**5 / (pi**1.5 * sig)) * np.exp(-(kr**2))

        Fx = front * (my * (-dz / r) + mz * (dy / r))

        Fy = front * (mx * (dz / r) + mz * (-dx / r))

        Fz = front * (mx * (-dy / r) + my * (dx / r))

    return Fx, Fy, Fz


def TransientElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, t, moment, fieldType="h", mu_r=1
):
    mu = 4 * np.pi * 1e-7 * mu_r

    if isinstance(moment, str):
        if moment.upper() == "X":
            mx, my, mz = 1.0, 0.0, 0.0
        elif moment.upper() == "Y":
            mx, my, mz = 0.0, 1.0, 0.0
        elif moment.upper() == "Z":
            mx, my, mz = 0.0, 0.0, 1.0
        else:
            raise NotImplementedError("String type for moment not recognized")

    else:
        mx, my, mz = moment[0], moment[1], moment[2]

    XYZ = utils.as_array_n_by_dim(XYZ, 3)

    dx = XYZ[:, 0] - srcLoc[0]
    dy = XYZ[:, 1] - srcLoc[1]
    dz = XYZ[:, 2] - srcLoc[2]

    r = np.sqrt(dx**2.0 + dy**2.0 + dz**2.0)
    k = np.sqrt(mu * sig / (4 * t))
    kr = k * r

    if fieldType == "e":
        front = 1 / (4.0 * np.pi * sig * r**3)
        mid = 3 * erf(kr) - (4 * kr**3 + 6 * kr) * np.exp(-(kr**2)) / np.sqrt(pi)
        end = -erf(kr) + (4 * kr**3 + 2 * kr) * np.exp(-(kr**2)) / np.sqrt(pi)

        Fx = front * (
            mx * ((dx**2 / r**2) * mid + end)
            + my * (dy * dx / r**2) * mid
            + mz * (dz * dx / r**2) * mid
        )

        Fy = front * (
            mx * (dx * dy / r**2) * mid
            + my * ((dy**2 / r**2) * mid + end)
            + mz * (dz * dy / r**2) * mid
        )

        Fz = front * (
            mx * (dx * dz / r**2) * mid
            + my * (dy * dz / r**2) * mid
            + mz * ((dz**2 / r**2) * mid + end)
        )

    elif fieldType == "h":
        front = (1 / (4.0 * pi * r**3)) * (
            erf(kr) - 2 * kr * np.exp(-(kr**2)) / np.sqrt(pi)
        )

        Fx = front * (my * -dz + mz * dy)

        Fy = front * (mx * dz + mz * -dx)

        Fz = front * (mx * -dy + my * dx)

    elif fieldType == "dhdt":
        front = -(2 * k**5 / (pi**1.5 * mu * sig)) * np.exp(-(kr**2))

        Fx = front * (my * -dz + mz * dy)

        Fy = front * (mx * dz + mz * -dx)

        Fz = front * (mx * -dy + my * dx)

    return Fx, Fy, Fz



from scipy.constants import G
from SimPEG.utils import mkvc
import numpy as np


def GravSphereFreeSpace(x, y, z, R, xc, yc, zc, rho):
    """
    Computing the gravity response of a sphere in free-space.
    >> Input
    x, y, z:   Observation locations
    R:     radius of the sphere
    xc, yc, zc: Location of the sphere
    rho: Density of sphere

    """
    if ~np.size(x) == np.size(y) == np.size(z):
        print("Specify same size of x, y, z")
        return

    unit_conv = 1e8  # Unit conversion from SI to (mgal*g/cc)
    x = mkvc(x)
    y = mkvc(y)
    z = mkvc(z)

    rx = x - xc
    ry = y - yc
    rz = z - zc

    r = np.sqrt((rx) ** 2 + (ry) ** 2 + (rz) ** 2)

    M = np.empty_like(x)  # create a vector of "Ms" if the point is outide
    M[r >= R] = R**3 * 4.0 / 3.0 * np.pi * rho  # outside points
    M[r < R] = r[r < R] ** 3 * 4.0 / 3.0 * np.pi * rho  # inside points

    g = -G * (1.0 / r**2) * M * unit_conv

    gx = g * (rx / r)
    gy = g * (ry / r)
    gz = g * (rz / r)

    return gx, gy, gz


def GravityGradientSphereFreeSpace(x, y, z, R, xc, yc, zc, rho):
    """
    Computing the induced response of magnetic sphere in free-space.
    >> Input
    x, y, z:   Observation locations
    R:     radius of the sphere
    xc, yc, zc: Location of the sphere
    rho: Density of sphere

    """
    if ~np.size(x) == np.size(y) == np.size(z):
        print("Specify same size of x, y, z")
        return

    unit_conv = 1e12  # Unit conversion from SI to (Eotvos)
    x = mkvc(x)
    y = mkvc(y)
    z = mkvc(z)

    rx = x - xc
    ry = y - yc
    rz = z - zc
    rx2 = rx * rx
    ry2 = ry * ry
    rz2 = rz * rz

    r = np.sqrt(rx2 + ry2 + rz2)
    bot = r * r * r * r * r

    M = np.empty_like(x)  # create a vector of "Ms" if the point is outide
    M[r >= R] = R**3 * 4.0 / 3.0 * np.pi * rho  # outside points
    M[r < R] = r[r < R] ** 3 * 4.0 / 3.0 * np.pi * rho  # inside points

    g = G * (1.0 / bot) * M * unit_conv

    gxx = g * (2 * rx2 - ry2 - rz2)
    gyy = g * (2 * ry2 - rx2 - rz2)
    gzz = -gxx - gyy
    gxy = g * (3 * rx * ry)
    gxz = g * (3 * rx * rz)
    gyz = g * (3 * ry * rz)

    return gxx, gxy, gxz, gyy, gyz, gzz



"""
=======================================================================
Gravity Simulation (:mod:`SimPEG.potential_fields.gravity`)
=======================================================================
.. currentmodule:: SimPEG.potential_fields.gravity

About ``gravity``

Simulations
===========
.. autosummary::
  :toctree: generated/

  Simulation3DIntegral
  SimulationEquivalentSourceLayer
  Simulation3DDifferential

Survey, Sources and Receivers
============================

.. autosummary::
  :toctree: generated/

  Point
  SourceField
  Survey


Analytic functions
==================
.. autosummary::
  :toctree: generated/

  analytics.GravSphereFreeSpace
  analytics.GravityGradientSphereFreeSpace

"""
from . import survey
from . import sources
from . import receivers
from . import analytics
from . import simulation

from .simulation import (
    Simulation3DIntegral,
    SimulationEquivalentSourceLayer,
    Simulation3DDifferential,
)
from .survey import Survey
from .sources import SourceField
from .receivers import Point



import numpy as np
import scipy.sparse as sp

import warnings
import uuid

from .utils import Counter
from .utils.code_utils import (
    validate_location_property,
    validate_ndarray_with_shape,
    validate_list_of_types,
    validate_type,
)


class BaseRx:
    """Base SimPEG receiver class.

    Parameters
    ----------
    locations : (n_loc, ndim) numpy.ndarray
        Locations assocated with a given receiver
    storeProjections : bool, Default = ``False``
        Store projections from the mesh to receiver
    """

    _Ps = None

    def __init__(self, locations, storeProjections=False, **kwargs):
        self.locations = locations

        # Deprecated properties
        rxType = kwargs.pop("rxType", None)
        if rxType is not None:
            raise AttributeError(
                "BaseRx no longer has an rxType property. Each receiver type is defined by "
                "a different receiver class."
            )
        projGLoc = kwargs.pop("projGLoc", None)
        if projGLoc is not None:
            warnings.warn(
                "'projGLoc' is no longer of property of the receiver class. It is set automatically "
                "based on the receiver and simulation class. Will be remove in SimPEG 0.18.0"
            )
        # ideally there shouldn't be any kwargs left to hit, but this will throw an
        # error if kwargs hasn't been emptied properly. This also will allow proper
        # initialization in multiple inheritence.
        super().__init__(**kwargs)

        # Remaining properties
        if getattr(self, "_Ps", None) is None:
            self._Ps = {}
        self.storeProjections = storeProjections
        self._uid = uuid.uuid4()

    @property
    def locations(self):
        """Receiver locations

        Returns
        -------
        (n_loc, n_dim) np.ndarray
            Receiver locations.
        """
        return self._locations

    @locations.setter
    def locations(self, locs):
        self._locations = validate_ndarray_with_shape(
            "locations", locs, shape=("*", "*"), dtype=float
        )

    # @property
    # def projected_grid(self):
    #     """Gridded locations being projected to receiver locations.

    #     A ``str`` is used to the define the projection from the gridded locations
    #     on the mesh to the receiver locations. The choices are as follows:

    #     - "CC"  --> cell centers
    #     - "Fx"  --> x-faces
    #     - "Fy"  --> y-faces
    #     - "Fz"  --> z-faces
    #     - "Ex"  --> x-edges
    #     - "Ey"  --> y-edges
    #     - "Ez"  --> z-edges
    #     - "N"   --> nodes

    #     Returns
    #     -------
    #     str
    #         The gridded locations being projected to the receiver locations.
    #     """
    #     return self._projected_grid

    # @projected_grid.setter
    # def projected_grid(self, var):
    #     if (var in ["CC", "Fx", "Fy", "Fz", "Ex", "Ey", "Ez", "N"]) == False:
    #         raise TypeError(
    #             "projected_grid must be one of the following strings: "
    #             "'CC', 'Fx', 'Fy', 'Fz', 'Ex', 'Ey', 'Ez', 'N'"
    #             f"Got {type(var)}"
    #         )

    #     self._projected_grid = var

    @property
    def uid(self):
        """Universal unique identifier

        Returns
        -------
        uuid.UUID
            A universal unique identifier
        """
        return self._uid

    @property
    def nD(self):
        """Number of data associated with the receiver

        Returns
        -------
        int
            Number of data associated with the receiver
        """
        return self.locations.shape[0]

    def getP(self, mesh, projected_grid="CC"):
        """Get projection matrix from mesh to receivers

        Parameters
        ----------
        mesh : discretize.BaseMesh
            A discretize mesh
        projected_grid : str
            Define what part of the mesh (i.e. edges, faces, centers, nodes) to
            project from. Must be one of::

                'Ex', 'edges_x'           -> x-component of field defined on x edges
                'Ey', 'edges_y'           -> y-component of field defined on y edges
                'Ez', 'edges_z'           -> z-component of field defined on z edges
                'Fx', 'faces_x'           -> x-component of field defined on x faces
                'Fy', 'faces_y'           -> y-component of field defined on y faces
                'Fz', 'faces_z'           -> z-component of field defined on z faces
                'N', 'nodes'              -> scalar field defined on nodes
                'CC', 'cell_centers'      -> scalar field defined on cell centers
                'CCVx', 'cell_centers_x'  -> x-component of vector field defined on cell centers
                'CCVy', 'cell_centers_y'  -> y-component of vector field defined on cell centers
                'CCVz', 'cell_centers_z'  -> z-component of vector field defined on cell centers

        Returns
        -------
        scipy.sparse.csr_matrix
            P, the interpolation matrix
        """
        if (mesh, projected_grid) in self._Ps:
            return self._Ps[(mesh, projected_grid)]

        P = mesh.get_interpolation_matrix(self.locations, projected_grid)
        if self.storeProjections:
            self._Ps[(mesh, projected_grid)] = P
        return P

    def eval(self, **kwargs):  # noqa: A003
        """Not implemented for BaseRx"""
        raise NotImplementedError(
            "the eval method for {} has not been implemented".format(self)
        )

    def evalDeriv(self, **kwargs):
        """Not implemented for BaseRx"""
        raise NotImplementedError(
            "the evalDeriv method for {} has not been implemented".format(self)
        )


class BaseTimeRx(BaseRx):
    """Base SimPEG receiver class for time-domain simulations

    Parameters
    ----------
    locations : (n, dim) array_like
        Receiver locations
    times : array_like
        Time channels
    """

    def __init__(self, locations, times, **kwargs):
        super().__init__(locations=locations, **kwargs)

        self.times = times

    @property
    def times(self):
        """Time channels for the receiver

        Returns
        -------
        numpy.ndarray
            Time channels for the receiver
        """
        return self._times

    @times.setter
    def times(self, value):
        self._times = validate_ndarray_with_shape(
            "times", value, shape=("*",), dtype=float
        )

    @property
    def nD(self):
        """Number of data associated with the receiver.

        Returns
        -------
        int
            Number of data associated with the receiver
        """
        return self.locations.shape[0] * len(self.times)

    def getSpatialP(self, mesh, projected_grid="CC"):
        """Returns the spatial projection matrix from mesh to receivers.

        Parameters
        ----------
        mesh : discretize.BaseMesh
            A ``discretize`` mesh; i.e. ``TensorMesh``, ``CylindricalMesh``, ``TreeMesh``
        projected_grid :

        Returns
        -------
        scipy.sparse.csr_matrix
            The projection matrix from the mesh (one of cell centers, nodes,
            edges, etc...) to the receivers. The returned quantity is not stored
            in memory. Instead, it is created on demand when needed.
        """
        return mesh.get_interpolation_matrix(self.locations, projected_grid)

    def getTimeP(self, time_mesh, projected_time_grid="N"):
        """Returns the time projection matrix from all time steps to receiver time channels.

        Parameters
        ----------
        time_mesh: 1D discretize.TensorMesh
            A 1D tensor mesh defining the time steps; either at cell centers or nodes.

        Returns
        -------
        scipy.sparse.csr_matrix
            The projection matrix from the mesh (one of cell centers, nodes,
            edges, etc...) to the receivers. The returned quantity is not stored
            in memory. Instead, it is created on demand when needed.
        """
        return time_mesh.get_interpolation_matrix(self.times, projected_time_grid)

    def getP(self, mesh, time_mesh, projected_grid="CC", projected_time_grid="N"):
        """
        Returns the projection matrices as a
        list for all components collected by
        the receivers.

        Notes
        -----
        Projection matrices are stored as a dictionary (mesh, timeMesh)
        if `storeProjections` is ``True``
        """
        if (mesh, time_mesh) in self._Ps:
            return self._Ps[(mesh, time_mesh)]

        Ps = self.getSpatialP(mesh, projected_grid)
        Pt = self.getTimeP(time_mesh, projected_time_grid)
        P = sp.kron(Pt, Ps)

        if self.storeProjections:
            self._Ps[(mesh, time_mesh)] = P

        return P


class BaseSrc:
    """Base SimPEG source class.

    Parameters
    ----------
    receiver_list : list of SimPEG.survey.BaseRx objects
        Sets the receivers associated with the source
    location : (n_dim) numpy.ndarray
        Location of the source
    """

    def __init__(self, receiver_list=None, location=None, **kwargs):
        self._location = None
        if receiver_list is None:
            receiver_list = []
        self.receiver_list = receiver_list

        if location is not None:
            self.location = location

        self._uid = uuid.uuid4()

    # add a listener to receiver_list to reset the __rxOrder.
    def __setattr__(self, name, item):
        super().__setattr__(name, item)
        if name == "receiver_list":
            self.__rxOrder = None

    @property
    def location(self):
        """Source location

        Returns
        -------
        (n_dim) np.ndarray
            Source location.
        """
        return self._location

    @location.setter
    def location(self, loc):
        self._location = validate_location_property("location", loc)

    @property
    def receiver_list(self):
        """List of receivers associated with the source

        Returns
        -------
        list of SimPEG.survey.BaseRx
            List of receivers associated with the source
        """
        return self._receiver_list

    @receiver_list.setter
    def receiver_list(self, new_list):
        self._receiver_list = validate_list_of_types(
            "receiver_list", new_list, BaseRx, ensure_unique=True
        )

    @property
    def _rxOrder(self):
        if getattr(self.__rxOrder, None) is None:
            self.__rxOrder = {}
            [
                self.__rxOrder.setdefault(rx.uid, ii)
                for ii, rx in enumerate(self.receiver_list)
            ]
        return self.__rxOrder

    @property
    def uid(self):
        """Universal unique identifier

        Returns
        -------
        uuid.UUID
            A universal unique identifier
        """
        return self._uid

    _fields_per_source = 1

    def get_receiver_indices(self, receivers):
        """Get indices for a subset of receivers within the source's receivers list.

        Parameters
        ----------
        receivers : list of SimPEG.survey.BaseRx
            A subset list of receivers within the source's receivers list

        Returns
        -------
        numpy.ndarray of int
            Indices for the subset receivers
        """

        if not isinstance(receivers, list):
            receivers = [receivers]
        for rx in receivers:
            if getattr(rx, "_uid", None) is None:
                raise KeyError("Receiver does not have a _uid: {0!s}".format(str(rx)))
        inds = list(map(lambda rx: self._rxOrder.get(rx._uid, None), receivers))
        if None in inds:
            raise KeyError(
                "Some of the receiver specified are not in this survey. "
                "{0!s}".format(str(inds))
            )
        return inds

    @property
    def nD(self):
        """Number of data associated with the source.

        Returns
        -------
        int
            Total number of data associated with the source.
        """
        return sum(self.vnD)

    @property
    def vnD(self):
        """Vector number of data.

        Returns
        -------
        np.ndarray of int
            Returns the corresponding number of data for each receiver.
        """
        return np.array([rx.nD for rx in self.receiver_list])


# TODO: allow a reciever list to be provided and assume it is used for all
# sources? (and store the projections)
class BaseSurvey:
    """Base SimPEG survey class.

    Parameters
    ----------
    source_list : list of SimPEG.survey.BaseSrc objects
        Sets the sources (and their receivers)
    counter : SimPEG.utils.Counter
        A SimPEG counter object
    """

    def __init__(self, source_list, counter=None, **kwargs):
        # Source list
        if source_list is None:
            source_list = []
        self.source_list = source_list

        if counter is not None:
            self.counter = counter

        self._uid = uuid.uuid4()
        super().__init__(**kwargs)

    @property
    def source_list(self):
        """List of sources associated with the survey

        Returns
        -------
        list of SimPEG.survey.BaseSrc
            List of sources associated with the survey
        """
        return self._source_list

    # add a listener to source_list to reset the __sourceOrder.
    def __setattr__(self, name, item):
        super().__setattr__(name, item)
        if name == "source_list":
            self.__sourceOrder = None

    @source_list.setter
    def source_list(self, new_list):
        self._source_list = validate_list_of_types(
            "source_list", new_list, BaseSrc, ensure_unique=True
        )

    @property
    def _sourceOrder(self):
        if getattr(self, "__sourceOrder", None) is None:
            self.__sourceOrder = {}
            ii = 0
            for src in self.source_list:
                n_fields = src._fields_per_source
                self.__sourceOrder[src._uid] = [ii + i for i in range(n_fields)]
                ii += n_fields
        return self.__sourceOrder

    @property
    def uid(self):
        """Universal unique identifier

        Returns
        -------
        uuid.UUID
            A universal unique identifier
        """
        return self._uid

    @property
    def counter(self):
        """A SimPEG counter object for counting iterations and operations

        Returns
        -------
        SimPEG.utils.counter_utils.Counter
            A SimPEG counter object
        """
        return self._counter

    @counter.setter
    def counter(self, new_obj):
        self._counter = validate_type("counter", new_obj, Counter, cast=False)

    # TODO: this should be private
    def get_source_indices(self, sources):
        if not isinstance(sources, list):
            sources = [sources]

        inds = []
        for src in sources:
            if getattr(src, "uid", None) is None:
                raise KeyError("Source does not have a _uid: {0!s}".format(str(src)))
            ind = self._sourceOrder.get(src.uid, None)
            if ind is None:
                raise KeyError(
                    "Some of the sources specified are not in this survey. "
                    "{0!s}".format(str(inds))
                )
            inds.extend(ind)
        return inds

    @property
    def nD(self):
        """Total number of data for the survey

        Returns
        -------
        int
            Total number of data for the survey
        """
        return sum(self.vnD)

    @property
    def vnD(self):
        """Number of associated data for each source

        Returns
        -------
        (n_src) numpy.ndarray of int
            Number of associate data for each source
        """
        if getattr(self, "_vnD", None) is None:
            self._vnD = np.array([src.nD for src in self.source_list])
        return self._vnD

    @property
    def nSrc(self):
        """Number of Sources

        Returns
        -------
        int
            Number of sources
        """
        return len(self.source_list)

    @property
    def _n_fields(self):
        """number of fields required for solution"""
        return sum(src._fields_per_source for src in self.source_list)


class BaseTimeSurvey(BaseSurvey):
    """Base SimPEG survey class for time-dependent simulations."""

    @property
    def unique_times(self):
        """Unique time channels for all survey receivers.

        Returns
        -------
        numpy.ndarray
            The unique time channels for all survey receivers.
        """
        if getattr(self, "_unique_times", None) is None:
            rx_times = []
            for source in self.source_list:
                for receiver in source.receiver_list:
                    rx_times.append(receiver.times)
            self._unique_times = np.unique(np.hstack(rx_times))
        return self._unique_times



import numpy as np
from scipy.constants import mu_0

from .... import Solver
from ....utils import sdiag

from .analytic_1d import getEHfields


def get1DEfields(m1d, sigma, freq, sourceAmp=1.0):
    """Function to get 1D electrical fields"""

    # Get the gradient
    G = m1d.nodal_gradient
    # Mass matrices
    # Magnetic permeability
    Mmu = sdiag(m1d.cell_volumes * (1.0 / mu_0))
    # Conductivity
    Msig = m1d.get_face_inner_product(sigma)
    # Set up the solution matrix
    A = G.T * Mmu * G + 1j * 2.0 * np.pi * freq * Msig
    # Define the inner part of the solution matrix
    Aii = A[1:-1, 1:-1]
    # Define the outer part of the solution matrix
    Aio = A[1:-1, [0, -1]]

    # Set the boundary conditions
    Ed, Eu, Hd, Hu = getEHfields(m1d, sigma, freq, m1d.nodes_x)
    Etot = Ed + Eu
    if sourceAmp is not None:
        Etot = (
            Etot / Etot[-1]
        ) * sourceAmp  # Scale the fields to be equal to sourceAmp at the top
    ## Note: The analytic solution is derived with e^iwt
    bc = np.r_[Etot[0], Etot[-1]]
    # The right hand side
    rhs = Aio * bc
    # Solve the system
    Aii_inv = Solver(Aii)
    eii = Aii_inv * rhs
    # Assign the boundary conditions
    e = np.r_[bc[0], eii, bc[1]]
    # Return the electrical fields
    return e



import numpy as np
from scipy.constants import epsilon_0, mu_0
from SimPEG import utils

omega = lambda f: 2.0 * np.pi * f
# TODO:
# r = lambda dx, dy, dz: np.sqrt( dx**2. + dy**2. + dz**2.)
# k = lambda f, mu, epsilon, sig: np.sqrt( omega(f)**2. *mu*epsilon -1j*omega(f)*mu*sig )


def E_from_ElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, f, current=1.0, length=1.0, orientation="X", kappa=0.0, epsr=1.0
):
    """
    Computing Analytic Electric fields from Electrical Dipole in a Wholespace
    TODO:
        Add description of parameters
    """
    mu = mu_0 * (1 + kappa)
    epsilon = epsilon_0 * epsr
    sig_hat = sig + 1j * omega(f) * epsilon

    XYZ = utils.as_array_n_by_dim(XYZ, 3)
    # Check
    if XYZ.shape[0] > 1 & f.shape[0] > 1:
        raise Exception(
            "I/O type error: For multiple field locations only a single frequency can be specified."
        )

    dx = XYZ[:, 0] - srcLoc[0]
    dy = XYZ[:, 1] - srcLoc[1]
    dz = XYZ[:, 2] - srcLoc[2]

    r = np.sqrt(dx**2.0 + dy**2.0 + dz**2.0)
    # k  = np.sqrt( -1j*2.*np.pi*f*mu*sig )
    k = np.sqrt(omega(f) ** 2.0 * mu * epsilon - 1j * omega(f) * mu * sig)

    front = current * length / (4.0 * np.pi * sig_hat * r**3) * np.exp(-1j * k * r)
    mid = -(k**2) * r**2 + 3 * 1j * k * r + 3

    if orientation.upper() == "X":
        Ex = front * ((dx**2 / r**2) * mid + (k**2 * r**2 - 1j * k * r - 1.0))
        Ey = front * (dx * dy / r**2) * mid
        Ez = front * (dx * dz / r**2) * mid
        return Ex, Ey, Ez

    elif orientation.upper() == "Y":
        #  x--> y, y--> z, z-->x
        Ey = front * ((dy**2 / r**2) * mid + (k**2 * r**2 - 1j * k * r - 1.0))
        Ez = front * (dy * dz / r**2) * mid
        Ex = front * (dy * dx / r**2) * mid
        return Ex, Ey, Ez

    elif orientation.upper() == "Z":
        # x --> z, y --> x, z --> y
        Ez = front * ((dz**2 / r**2) * mid + (k**2 * r**2 - 1j * k * r - 1.0))
        Ex = front * (dz * dx / r**2) * mid
        Ey = front * (dz * dy / r**2) * mid
        return Ex, Ey, Ez


def E_galvanic_from_ElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, f, current=1.0, length=1.0, orientation="X", kappa=1.0, epsr=1.0
):
    """
    Computing Galvanic portion of Electric fields from Electrical Dipole in a Wholespace
    TODO:
        Add description of parameters
    """
    mu = mu_0 * (1 + kappa)
    epsilon = epsilon_0 * epsr
    sig_hat = sig + 1j * omega(f) * epsilon

    XYZ = utils.as_array_n_by_dim(XYZ, 3)
    # Check
    if XYZ.shape[0] > 1 & f.shape[0] > 1:
        raise Exception(
            "I/O type error: For multiple field locations only a single frequency can be specified."
        )

    dx = XYZ[:, 0] - srcLoc[0]
    dy = XYZ[:, 1] - srcLoc[1]
    dz = XYZ[:, 2] - srcLoc[2]

    r = np.sqrt(dx**2.0 + dy**2.0 + dz**2.0)
    # k  = np.sqrt( -1j*2.*np.pi*f*mu*sig )
    k = np.sqrt(omega(f) ** 2.0 * mu * epsilon - 1j * omega(f) * mu * sig)

    front = current * length / (4.0 * np.pi * sig_hat * r**3) * np.exp(-1j * k * r)
    mid = -(k**2) * r**2 + 3 * 1j * k * r + 3

    if orientation.upper() == "X":
        Ex_galvanic = front * ((dx**2 / r**2) * mid + (-1j * k * r - 1.0))
        Ey_galvanic = front * (dx * dy / r**2) * mid
        Ez_galvanic = front * (dx * dz / r**2) * mid
        return Ex_galvanic, Ey_galvanic, Ez_galvanic

    elif orientation.upper() == "Y":
        #  x--> y, y--> z, z-->x
        Ey_galvanic = front * ((dy**2 / r**2) * mid + (-1j * k * r - 1.0))
        Ez_galvanic = front * (dy * dz / r**2) * mid
        Ex_galvanic = front * (dy * dx / r**2) * mid
        return Ex_galvanic, Ey_galvanic, Ez_galvanic

    elif orientation.upper() == "Z":
        # x --> z, y --> x, z --> y
        Ez_galvanic = front * ((dz**2 / r**2) * mid + (-1j * k * r - 1.0))
        Ex_galvanic = front * (dz * dx / r**2) * mid
        Ey_galvanic = front * (dz * dy / r**2) * mid
        return Ex_galvanic, Ey_galvanic, Ez_galvanic


def E_inductive_from_ElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, f, current=1.0, length=1.0, orientation="X", kappa=1.0, epsr=1.0
):
    """
    Computing Inductive portion of Electric fields from Electrical Dipole in a Wholespace
    TODO:
        Add description of parameters
    """
    mu = mu_0 * (1 + kappa)
    epsilon = epsilon_0 * epsr
    sig_hat = sig + 1j * omega(f) * epsilon

    XYZ = utils.as_array_n_by_dim(XYZ, 3)
    # Check
    if XYZ.shape[0] > 1 & f.shape[0] > 1:
        raise Exception(
            "I/O type error: For multiple field locations only a single frequency can be specified."
        )

    dx = XYZ[:, 0] - srcLoc[0]
    dy = XYZ[:, 1] - srcLoc[1]
    dz = XYZ[:, 2] - srcLoc[2]

    r = np.sqrt(dx**2.0 + dy**2.0 + dz**2.0)
    # k  = np.sqrt( -1j*2.*np.pi*f*mu*sig )
    k = np.sqrt(omega(f) ** 2.0 * mu * epsilon - 1j * omega(f) * mu * sig)

    front = current * length / (4.0 * np.pi * sig_hat * r**3) * np.exp(-1j * k * r)

    if orientation.upper() == "X":
        Ex_inductive = front * (k**2 * r**2)
        Ey_inductive = np.zeros_like(Ex_inductive)
        Ez_inductive = np.zeros_like(Ex_inductive)
        return Ex_inductive, Ey_inductive, Ez_inductive

    elif orientation.upper() == "Y":
        #  x--> y, y--> z, z-->x
        Ey_inductive = front * (k**2 * r**2)
        Ez_inductive = np.zeros_like(Ey_inductive)
        Ex_inductive = np.zeros_like(Ey_inductive)
        return Ex_inductive, Ey_inductive, Ez_inductive

    elif orientation.upper() == "Z":
        # x --> z, y --> x, z --> y
        Ez_inductive = front * (k**2 * r**2)
        Ex_inductive = np.zeros_like(Ez_inductive)
        Ey_inductive = np.zeros_like(Ez_inductive)
        return Ex_inductive, Ey_inductive, Ez_inductive


def J_from_ElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, f, current=1.0, length=1.0, orientation="X", kappa=1.0, epsr=1.0
):
    """
    Computing Current densities from Electrical Dipole in a Wholespace
    TODO:
        Add description of parameters
    """

    Ex, Ey, Ez = E_from_ElectricDipoleWholeSpace(
        XYZ,
        srcLoc,
        sig,
        f,
        current=current,
        length=length,
        orientation=orientation,
        kappa=kappa,
        epsr=epsr,
    )
    Jx = sig * Ex
    Jy = sig * Ey
    Jz = sig * Ez
    return Jx, Jy, Jz


def J_galvanic_from_ElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, f, current=1.0, length=1.0, orientation="X", kappa=1.0, epsr=1.0
):
    """
    Computing Galvanic portion of Current densities from Electrical Dipole in a Wholespace
    TODO:
        Add description of parameters
    """

    Ex_galvanic, Ey_galvanic, Ez_galvanic = E_galvanic_from_ElectricDipoleWholeSpaced(
        XYZ,
        srcLoc,
        sig,
        f,
        current=current,
        length=length,
        orientation=orientation,
        kappa=kappa,
        epsr=epsr,
    )
    Jx_galvanic = sig * Ex_galvanic
    Jy_galvanic = sig * Ey_galvanic
    Jz_galvanic = sig * Ez_galvanic
    return Jx_galvanic, Jy_galvanic, Jz_galvanic


def J_inductive_from_ElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, f, current=1.0, length=1.0, orientation="X", kappa=1.0, epsr=1.0
):
    """
    Computing Inductive portion of Current densities from Electrical Dipole in a Wholespace
    TODO:
        Add description of parameters
    """

    (
        Ex_inductive,
        Ey_inductive,
        Ez_inductive,
    ) = E_inductive_from_ElectricDipoleWholeSpaced(
        XYZ,
        srcLoc,
        sig,
        f,
        current=current,
        length=length,
        orientation=orientation,
        kappa=kappa,
        epsr=epsr,
    )
    Jx_inductive = sig * Ex_inductive
    Jy_inductive = sig * Ey_inductive
    Jz_inductive = sig * Ez_inductive
    return Jx_inductive, Jy_inductive, Jz_inductive


def H_from_ElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, f, current=1.0, length=1.0, orientation="X", kappa=1.0, epsr=1.0
):
    """
    Computing Magnetic fields from Electrical Dipole in a Wholespace
    TODO:
        Add description of parameters
    """
    mu = mu_0 * (1 + kappa)
    epsilon = epsilon_0 * epsr
    XYZ = utils.as_array_n_by_dim(XYZ, 3)
    # Check
    if XYZ.shape[0] > 1 & f.shape[0] > 1:
        raise Exception(
            "I/O type error: For multiple field locations only a single frequency can be specified."
        )

    dx = XYZ[:, 0] - srcLoc[0]
    dy = XYZ[:, 1] - srcLoc[1]
    dz = XYZ[:, 2] - srcLoc[2]

    r = np.sqrt(dx**2.0 + dy**2.0 + dz**2.0)
    # k  = np.sqrt( -1j*2.*np.pi*f*mu*sig )
    k = np.sqrt(omega(f) ** 2.0 * mu * epsilon - 1j * omega(f) * mu * sig)

    front = (
        current
        * length
        / (4.0 * np.pi * r**2)
        * (-1j * k * r + 1)
        * np.exp(-1j * k * r)
    )

    if orientation.upper() == "X":
        Hy = front * (-dz / r)
        Hz = front * (dy / r)
        Hx = np.zeros_like(Hy)
        return Hx, Hy, Hz

    elif orientation.upper() == "Y":
        Hx = front * (dz / r)
        Hz = front * (-dx / r)
        Hy = np.zeros_like(Hx)
        return Hx, Hy, Hz

    elif orientation.upper() == "Z":
        Hx = front * (-dy / r)
        Hy = front * (dx / r)
        Hz = np.zeros_like(Hx)
        return Hx, Hy, Hz


def B_from_ElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, f, current=1.0, length=1.0, orientation="X", kappa=1.0, epsr=1.0
):
    """
    Computing Magnetic flux densites from Electrical Dipole in a Wholespace
    TODO:
        Add description of parameters
    """

    Hx, Hy, Hz = H_from_ElectricDipoleWholeSpace(
        XYZ,
        srcLoc,
        sig,
        f,
        current=current,
        length=length,
        orientation=orientation,
        kappa=kappa,
        epsr=epsr,
    )
    Bx = mu * Hx
    By = mu * Hy
    Bz = mu * Hz
    return Bx, By, Bz


def A_from_ElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, f, current=1.0, length=1.0, orientation="X", kappa=1.0, epsr=1.0
):
    """
    Computing Electric vector potentials from Electrical Dipole in a Wholespace
    TODO:
        Add description of parameters
    """
    mu = mu_0 * (1 + kappa)
    epsilon = epsilon_0 * epsr
    XYZ = utils.as_array_n_by_dim(XYZ, 3)
    # Check
    if XYZ.shape[0] > 1 & f.shape[0] > 1:
        raise Exception(
            "I/O type error: For multiple field locations only a single frequency can be specified."
        )

    dx = XYZ[:, 0] - srcLoc[0]
    dy = XYZ[:, 1] - srcLoc[1]
    dz = XYZ[:, 2] - srcLoc[2]

    r = np.sqrt(dx**2.0 + dy**2.0 + dz**2.0)
    k = np.sqrt(omega(f) ** 2.0 * mu * epsilon - 1j * omega(f) * mu * sig)

    front = current * length / (4.0 * np.pi * r)

    if orientation.upper() == "X":
        Ax = front * np.exp(-1j * k * r)
        Ay = np.zeros_like(Ax)
        Az = np.zeros_like(Ax)
        return Ax, Ay, Az

    elif orientation.upper() == "Y":
        Ay = front * np.exp(-1j * k * r)
        Ax = np.zeros_like(Ay)
        Az = np.zeros_like(Ay)
        return Ax, Ay, Az

    elif orientation.upper() == "Z":
        Az = front * np.exp(-1j * k * r)
        Ax = np.zeros_like(Ay)
        Ay = np.zeros_like(Ay)
        return Ax, Ay, Az



import dask
import dask.array
import os
from ....electromagnetics.time_domain.simulation import BaseTDEMSimulation as Sim
from ....utils import Zero
from SimPEG.fields import TimeFields
from multiprocessing import cpu_count
import numpy as np
import scipy.sparse as sp
from dask import array, delayed

from SimPEG.dask.simulation import dask_Jvec, dask_Jtvec, dask_getJtJdiag
from SimPEG.dask.utils import get_parallel_blocks
from SimPEG.utils import mkvc
import zarr
from time import time
from tqdm import tqdm

Sim.sensitivity_path = "./sensitivity/"
Sim.store_sensitivities = "ram"

Sim.getJtJdiag = dask_getJtJdiag
Sim.Jvec = dask_Jvec
Sim.Jtvec = dask_Jtvec
Sim.clean_on_model_update = ["_Jmatrix", "_jtjdiag"]


@delayed
def field_projection(field_array, src_list, array_ind, time_ind, func):
    fieldI = field_array[:, :, array_ind]
    if fieldI.shape[0] == fieldI.size:
        fieldI = mkvc(fieldI, 2)
    new_array = func(fieldI, src_list, time_ind)
    if new_array.ndim == 1:
        new_array = new_array[:, np.newaxis, np.newaxis]
    elif new_array.ndim == 2:
        new_array = new_array[:, :, np.newaxis]

    return new_array


def _getField(self, name, ind, src_list):
    srcInd, timeInd = ind

    if name in self._fields:
        out = self._fields[name][:, srcInd, timeInd]
    else:
        # Aliased fields
        alias, loc, func = self.aliasFields[name]
        if isinstance(func, str):
            assert hasattr(self, func), (
                "The alias field function is a string, but it does "
                "not exist in the Fields class."
            )
            func = getattr(self, func)
        pointerFields = self._fields[alias][:, srcInd, timeInd]
        pointerShape = self._correctShape(alias, ind)
        pointerFields = pointerFields.reshape(pointerShape, order="F")

        # First try to return the function as three arguments (without timeInd)
        if timeInd == slice(None, None, None):
            try:
                # assume it will take care of integrating over all times
                return func(pointerFields, srcInd)
            except TypeError:
                pass

        timeII = np.arange(self.simulation.nT + 1)[timeInd]
        if not isinstance(src_list, list):
            src_list = [src_list]

        if timeII.size == 1:
            pointerShapeDeflated = self._correctShape(alias, ind, deflate=True)
            pointerFields = pointerFields.reshape(pointerShapeDeflated, order="F")
            out = func(pointerFields, src_list, timeII)
        else:  # loop over the time steps
            nT = pointerShape[2]
            arrays = []

            for i, TIND_i in enumerate(timeII):  # Need to parallelize this
                arrays.append(
                    array.from_delayed(
                        field_projection(pointerFields, src_list, i, TIND_i, func),
                        dtype=np.float32,
                        shape=(pointerShape[0], pointerShape[1], 1),
                    )
                )

            out = array.dstack(arrays).compute()

    shape = self._correctShape(name, ind, deflate=True)
    return out.reshape(shape, order="F")


TimeFields._getField = _getField


def fields(self, m=None, return_Ainv=False):
    if m is not None:
        self.model = m

    f = self.fieldsPair(self)
    f[:, self._fieldType + "Solution", 0] = self.getInitialFields()
    Ainv = {}
    ATinv = {}

    for tInd, dt in enumerate(self.time_steps):
        if dt not in Ainv:
            A = self.getAdiag(tInd)
            Ainv[dt] = self.solver(sp.csr_matrix(A), **self.solver_opts)
            if return_Ainv:
                ATinv[dt] = self.solver(sp.csr_matrix(A.T), **self.solver_opts)

        Asubdiag = self.getAsubdiag(tInd)
        rhs = -Asubdiag * f[:, (self._fieldType + "Solution"), tInd]

        if (
            np.abs(self.survey.source_list[0].waveform.eval(self.times[tInd + 1]))
            > 1e-8
        ):
            rhs += self.getRHS(tInd + 1)

        sol = Ainv[dt] * rhs
        f[:, self._fieldType + "Solution", tInd + 1] = sol

    for A in Ainv.values():
        A.clean()

    if return_Ainv:
        return f, ATinv
    else:
        return f, None


Sim.fields = fields


def dask_getSourceTerm(self, tInd):
    """
    Assemble the source term. This ensures that the RHS is a vector / array
    of the correct size
    """
    source_list = self.survey.source_list
    source_block = np.array_split(source_list, cpu_count())

    def source_evaluation(simulation, sources, time):
        s_m, s_e = [], []
        for source in sources:
            sm, se = source.eval(simulation, time)
            s_m.append(sm)
            s_e.append(se)

        return s_m, s_e

    block_compute = []
    for block in source_block:
        block_compute.append(
            delayed(source_evaluation, pure=True)(self, block, self.times[tInd])
        )

    eval = dask.compute(block_compute)[0]

    s_m, s_e = [], []
    for block in eval:
        if block[0]:
            s_m.append(block[0])
            s_e.append(block[1])

    if isinstance(s_m[0][0], Zero):
        return Zero(), np.vstack(s_e).T

    return np.vstack(s_m).T, np.vstack(s_e).T


Sim.getSourceTerm = dask_getSourceTerm


@delayed
def evaluate_receivers(block, mesh, time_mesh, fields, fields_array):
    data = []
    for source, ind, receiver in block:
        Ps = receiver.getSpatialP(mesh, fields)
        Pt = receiver.getTimeP(time_mesh, fields)
        vector = (Pt * (Ps * fields_array[:, ind, :]).T).flatten()

        data.append(vector)

    return np.hstack(data)


def dask_dpred(self, m=None, f=None, compute_J=False):
    """
    dpred(m, f=None)
    Create the projected data from a model.
    The fields, f, (if provided) will be used for the predicted data
    instead of recalculating the fields (which may be expensive!).

    .. math::

        d_\\text{pred} = P(f(m))

    Where P is a projection of the fields onto the data space.
    """
    if self.survey is None:
        raise AttributeError(
            "The survey has not yet been set and is required to compute "
            "data. Please set the survey for the simulation: "
            "simulation.survey = survey"
        )

    if f is None:
        if m is None:
            m = self.model
        f, Ainv = self.fields(m, return_Ainv=compute_J)

    rows = []
    receiver_projection = self.survey.source_list[0].receiver_list[0].projField
    fields_array = f[:, receiver_projection, :]

    if len(self.survey.source_list) == 1:
        fields_array = fields_array[:, np.newaxis, :]

    all_receivers = []

    for ind, src in enumerate(self.survey.source_list):
        for rx in src.receiver_list:
            all_receivers.append((src, ind, rx))

    receiver_blocks = np.array_split(all_receivers, cpu_count())

    for block in receiver_blocks:
        n_data = np.sum(rec.nD for _, _, rec in block)
        if n_data == 0:
            continue

        rows.append(
            array.from_delayed(
                evaluate_receivers(block, self.mesh, self.time_mesh, f, fields_array),
                dtype=np.float64,
                shape=(n_data,),
            )
        )

    data = array.hstack(rows).compute()

    if compute_J and self._Jmatrix is None:
        Jmatrix = self.compute_J(f=f, Ainv=Ainv)
        return data, Jmatrix

    return data


Sim.dpred = dask_dpred
Sim.field_derivs = None


@delayed
def delayed_block_deriv(
    n_times, chunks, field_len, source_list, mesh, time_mesh, fields, shape
):
    """Compute derivatives for sources and receivers in a block"""
    df_duT = []
    j_updates = []

    for indices, arrays in chunks:
        j_update = 0.0
        source = source_list[indices[0]]
        receiver = source.receiver_list[indices[1]]

        spatialP = receiver.getSpatialP(mesh, fields)
        timeP = receiver.getTimeP(time_mesh, fields)

        derivative_fun = getattr(fields, "_{}Deriv".format(receiver.projField), None)
        time_derivs = []
        for time_index in range(n_times + 1):
            if len(timeP[:, time_index].data) == 0:
                time_derivs.append(
                    sp.csr_matrix((field_len, len(arrays[0])), dtype=np.float32)
                )
                j_update += sp.csr_matrix((arrays[0].shape[0], shape), dtype=np.float32)
                continue

            projection = sp.kron(timeP[:, time_index], spatialP, format="csr")
            cur = derivative_fun(
                time_index,
                source,
                None,
                projection.T,
                adjoint=True,
            )

            time_derivs.append(cur[0][:, arrays[0]])

            if not isinstance(cur[1], Zero):
                j_update += cur[1].T
            else:
                j_update += sp.csr_matrix((arrays[0].shape[0], shape), dtype=np.float32)

        j_updates.append(j_update)
        df_duT.append(time_derivs)

    return df_duT, j_updates


def compute_field_derivs(simulation, fields, blocks, Jmatrix, fields_shape):
    """
    Compute the derivative of the fields
    """
    delayed_chunks = []
    for chunks in blocks:
        if len(chunks) == 0:
            continue

        delayed_block = delayed_block_deriv(
            simulation.nT,
            chunks,
            fields_shape[0],
            simulation.survey.source_list,
            simulation.mesh,
            simulation.time_mesh,
            fields,
            simulation.model.size,
        )
        delayed_chunks.append(delayed_block)

    result = dask.compute(delayed_chunks)[0]
    df_duT = [
        [[[] for _ in block] for block in blocks if len(block) > 0]
        for _ in range(simulation.nT + 1)
    ]
    j_updates = []

    for bb, block in enumerate(result):
        j_updates += block[1]
        for cc, chunk in enumerate(block[0]):
            for ind, time_block in enumerate(chunk):
                df_duT[ind][bb][cc] = time_block

    j_updates = sp.vstack(j_updates)

    if len(j_updates.data) > 0:
        Jmatrix += j_updates
        if simulation.store_sensitivities == "disk":
            sens_name = simulation.sensitivity_path[:-5] + f"_{time_index % 2}.zarr"
            array.to_zarr(Jmatrix, sens_name, compute=True, overwrite=True)
            Jmatrix = array.from_zarr(sens_name)

    return df_duT, Jmatrix


@delayed
def deriv_block(
    s_id, r_id, b_id, ATinv_df_duT_v, Asubdiag, local_ind, field_derivs, tInd
):
    if (s_id, r_id, b_id) not in ATinv_df_duT_v:
        # last timestep (first to be solved)
        stacked_block = field_derivs.toarray()[:, local_ind]

    else:
        stacked_block = np.asarray(
            field_derivs[:, local_ind]
            - Asubdiag.T * ATinv_df_duT_v[(s_id, r_id, b_id)][:, local_ind]
        )

    return stacked_block


def update_deriv_blocks(address, indices, derivatives, solve, shape):
    if address not in derivatives:
        deriv_array = np.zeros(shape)
    else:
        deriv_array = derivatives[address]

    if address in indices:
        columns, local_ind = indices[address]
        if solve is not None:
            deriv_array[:, local_ind] = solve[:, columns]

    derivatives[address] = deriv_array


def get_field_deriv_block(
    simulation,
    block: list,
    field_derivs: list,
    tInd: int,
    AdiagTinv,
    ATinv_df_duT_v: dict,
    time_mask,
):
    """
    Stack the blocks of field derivatives for a given timestep and call the direct solver.
    """
    stacked_blocks = []
    indices = {}
    count = 0

    Asubdiag = None
    if tInd < simulation.nT - 1:
        Asubdiag = simulation.getAsubdiag(tInd + 1)

    for ((s_id, r_id, b_id), (rx_ind, j_ind, shape)), field_deriv in zip(
        block, field_derivs
    ):
        # Cut out early data
        time_check = np.kron(time_mask, np.ones(shape, dtype=bool))[rx_ind]
        local_ind = np.arange(rx_ind.shape[0])[time_check]

        if len(local_ind) < 1:
            continue

        indices[(s_id, r_id, b_id)] = (
            np.arange(count, count + len(local_ind)),
            local_ind,
        )
        count += len(local_ind)
        deriv_comp = deriv_block(
            s_id,
            r_id,
            b_id,
            ATinv_df_duT_v,
            Asubdiag,
            local_ind,
            field_deriv,
            tInd,
        )
        stacked_blocks.append(
            array.from_delayed(
                deriv_comp,
                dtype=float,
                shape=(
                    field_deriv.shape[0],
                    len(local_ind),
                ),
            )
        )
    if len(stacked_blocks) > 0:
        blocks = array.hstack(stacked_blocks).compute()
        solve = (AdiagTinv * blocks).reshape(blocks.shape)
    else:
        solve = None

    update_list = []
    for (address, arrays), field_deriv in zip(block, field_derivs):
        shape = (
            field_deriv.shape[0],
            len(arrays[0]),
        )

        update_deriv_blocks(address, indices, ATinv_df_duT_v, solve, shape)

    return ATinv_df_duT_v


@delayed
def compute_rows(
    simulation,
    tInd,
    chunks,
    ATinv_df_duT_v,
    fields,
    time_mask,
):
    """
    Compute the rows of the sensitivity matrix for a given source and receiver.
    """
    n_rows = np.sum(len(chunk[1][0]) for chunk in chunks)
    rows = []

    for address, ind_array in chunks:
        src = simulation.survey.source_list[address[0]]
        time_check = np.kron(time_mask, np.ones(ind_array[2], dtype=bool))[ind_array[0]]
        local_ind = np.arange(len(ind_array[0]))[time_check]

        if len(local_ind) < 1:
            row_block = np.zeros(
                (len(ind_array[1]), simulation.model.size), dtype=np.float32
            )
            rows.append(row_block)
            continue

        field_derivs = ATinv_df_duT_v[address]
        dAsubdiagT_dm_v = simulation.getAsubdiagDeriv(
            tInd,
            fields[:, address[0], tInd],
            field_derivs[:, local_ind],
            adjoint=True,
        )

        dRHST_dm_v = simulation.getRHSDeriv(
            tInd + 1, src, field_derivs[:, local_ind], adjoint=True
        )  # on nodes of time mesh

        un_src = fields[:, address[0], tInd + 1]
        # cell centered on time mesh
        dAT_dm_v = simulation.getAdiagDeriv(
            tInd, un_src, field_derivs[:, local_ind], adjoint=True
        )
        row_block = np.zeros(
            (len(ind_array[1]), simulation.model.size), dtype=np.float32
        )
        row_block[time_check, :] = (-dAT_dm_v - dAsubdiagT_dm_v + dRHST_dm_v).T.astype(
            np.float32
        )
        rows.append(row_block)

    return np.vstack(rows)


def compute_J(self, f=None, Ainv=None):
    """
    Compute the rows for the sensitivity matrix.
    """
    if f is None:
        f, Ainv = self.fields(self.model, return_Ainv=True)

    ftype = self._fieldType + "Solution"
    sens_name = self.sensitivity_path[:-5]
    if self.store_sensitivities == "disk":
        rows = array.zeros(
            (self.survey.nD, self.model.size),
            chunks=(self.max_chunk_size, self.model.size),
            dtype=np.float32,
        )
        Jmatrix = array.to_zarr(
            rows,
            os.path.join(sens_name + "_1.zarr"),
            compute=True,
            return_stored=True,
            overwrite=True,
        )
    else:
        Jmatrix = np.zeros((self.survey.nD, self.model.size), dtype=np.float64)

    simulation_times = np.r_[0, np.cumsum(self.time_steps)] + self.t0
    data_times = self.survey.source_list[0].receiver_list[0].times
    blocks = get_parallel_blocks(
        self.survey.source_list, self.model.shape[0], self.max_chunk_size
    )
    fields_array = f[:, ftype, :]

    if len(self.survey.source_list) == 1:
        fields_array = fields_array[:, np.newaxis, :]

    times_field_derivs, Jmatrix = compute_field_derivs(
        self, f, blocks, Jmatrix, fields_array.shape
    )

    ATinv_df_duT_v = {}
    for tInd, dt in tqdm(zip(reversed(range(self.nT)), reversed(self.time_steps))):
        AdiagTinv = Ainv[dt]
        j_row_updates = []
        time_mask = data_times > simulation_times[tInd]

        if not np.any(time_mask):
            continue

        tc = time()
        for block, field_deriv in zip(blocks, times_field_derivs[tInd + 1]):
            ATinv_df_duT_v = get_field_deriv_block(
                self, block, field_deriv, tInd, AdiagTinv, ATinv_df_duT_v, time_mask
            )

            if len(block) == 0:
                continue

            j_row_updates.append(
                array.from_delayed(
                    compute_rows(
                        self,
                        tInd,
                        block,
                        ATinv_df_duT_v,
                        fields_array,
                        time_mask,
                    ),
                    dtype=np.float32,
                    shape=(
                        np.sum(len(chunk[1][0]) for chunk in block),
                        self.model.size,
                    ),
                )
            )

        if self.store_sensitivities == "disk":
            sens_name = self.sensitivity_path[:-5] + f"_{tInd % 2}.zarr"
            array.to_zarr(
                Jmatrix + array.vstack(j_row_updates),
                sens_name,
                compute=True,
                overwrite=True,
            )
            Jmatrix = array.from_zarr(sens_name)
        else:
            Jmatrix += array.vstack(j_row_updates).compute()

    for A in Ainv.values():
        A.clean()

    if self.store_sensitivities == "ram":
        return np.asarray(Jmatrix)

    return Jmatrix


Sim.compute_J = compute_J



import numpy as np
import scipy.sparse as sp
import matplotlib.pyplot as plt

from ...simulation import LinearSimulation
from ...utils import sub2ind
from ... import props


def _lengthInCell(O, D, x, y, plotIt=False):
    maxD = np.sqrt(np.sum(D**2))
    D = D / maxD

    def dist(a):
        return O + a * D

    if plotIt:
        plt.plot(x[[0, 0, 1, 1, 0]], y[[0, 1, 1, 0, 0]], "b")
        plt.plot(O[0], O[1], "rs")
        d = np.r_[0, maxD]
        plt.plot(O[0] + d * D[0], O[1] + d * D[1], "k-")

    alp = np.array([-np.inf, np.inf, -np.inf, np.inf])

    if np.abs(D[0]) > 0:
        alp[0] = (x[0] - O[0]) / D[0]
        alp[1] = (x[1] - O[0]) / D[0]
        if plotIt:
            plt.plot(dist(alp[0])[0], dist(alp[0])[1], "mo")
            plt.plot(dist(alp[1])[0], dist(alp[1])[1], "mo")

    if np.abs(D[1]) > 0:
        alp[2] = (y[0] - O[1]) / D[1]
        alp[3] = (y[1] - O[1]) / D[1]
        if plotIt:
            plt.plot(dist(alp[2])[0], dist(alp[2])[1], "go")
            plt.plot(dist(alp[3])[0], dist(alp[3])[1], "go")

    midAlp = np.array(sorted(alp)[1:3])
    midAlp[midAlp < 0] = 0
    midAlp[midAlp > maxD] = maxD
    midPoint = dist(np.mean(midAlp))

    #     print alp, midAlp, midPoint

    if (
        midPoint[0] >= x[0]
        and midPoint[0] <= x[1]
        and midPoint[1] >= y[0]
        and midPoint[1] <= y[1]
    ):
        vec = dist(midAlp[0]) - dist(midAlp[1])
        if plotIt:
            c = np.c_[dist(midAlp[0]), dist(midAlp[1])]
            plt.plot(c[0, :], c[1, :], "r", lw=2)
        return np.sqrt(np.sum(vec**2))

    return None


def _lineintegral(M, Tx, Rx):
    O, D = Tx, Rx - Tx
    I, J, V = [], [], []
    for i in range(M.shape_cells[0]):
        for j in range(M.shape_cells[1]):
            x = M.nodes_x[[i, i + 1]]
            y = M.nodes_y[[j, j + 1]]
            v = _lengthInCell(O, D, x, y)
            if v is not None:
                I += [i]
                J += [j]
                V += [v]
    inds = sub2ind(M.vnC, np.array([I, J]).T)
    return inds, V


class Simulation2DIntegral(LinearSimulation):
    slowness, slownessMap, slownessDeriv = props.Invertible("Slowness model (1/v)")

    def __init__(
        self, mesh=None, survey=None, slowness=None, slownessMap=None, **kwargs
    ):
        super().__init__(mesh=mesh, survey=survey, **kwargs)
        self.slowness = slowness
        self.slownessMap = slownessMap

    @property
    def A(self):
        if getattr(self, "_A", None) is not None:
            return self._A

        self._A = sp.lil_matrix((self.survey.nD, self.mesh.nC))
        row = 0
        for src in self.survey.source_list:
            for rx in src.receiver_list:
                for loc_i in range(rx.locations.shape[0]):
                    inds, V = _lineintegral(
                        self.mesh, src.location, rx.locations[loc_i, :]
                    )
                    self._A[inds * 0 + row, inds] = V
                    row += 1

        return self._A

    def fields(self, m):
        self.model = m
        return self.A * self.slowness

    def Jvec(self, m, v, f=None):
        self.model = m
        # mt = self.model.transformDeriv
        # return self.A * ( mt * v )
        return self.A * self.slownessDeriv * v

    def Jtvec(self, m, v, f=None):
        self.model = m
        # mt = self.model.transformDeriv
        # return mt.T * ( self.A.T * v )
        return self.slownessDeriv.T * self.A.T * v



import scipy.sparse as sp
import numpy as np

from .... import maps, props
from ....utils import validate_list_of_types, validate_active_indices
from .. import resistivity as dc
from .sources import StreamingCurrents


class Simulation3DCellCentered(dc.Simulation3DCellCentered):
    r"""A Spontaneous potential simulation.

    Parameters
    ----------
    mesh : discretize.base.BaseMesh
    survey : spontaneous_potential.Survey
    sigma, rho : float or array_like
        The conductivity/resistivity model of the subsurface.
    q : float, array_like, optional
        The charge density accumulation rate model (C/(s m^3)), also
        physically represents the volumetric current density (A/m^3).
    qMap : SimPEG.maps.IdentityMap, optional
        The mapping used to go from the simulation model to `q`. Set this
        to invert for `q`.
    **kwargs
        arguments passed on to :class:`.resistivity.Simulation3DCellCentered`

    Notes
    -----
    The charge density accumulation rate, :math:`q`, is related to the spontaneous
    electric potential, :math:`\phi`, with the same PDE, that relates current
    sources to potential in the resistivity case.

    .. math:: - \nabla \cdot \sigma \nabla \phi = q

    This equation is solve for potential with a finite volume approach,
    discretized with :math:`\phi` and :math:`q` on cell centers, electrical
    conductivity :math`\sigma` as a cell property, and therefore current density
    lives on the faces between cells.

    By default the boundary conditions assume a Robin condition on the subsurface
    boundaries, and a zero Nuemann boundary at the top. For more details on the
    boundary conditions, check out the resistivity simulations.
    """

    q, qMap, qDeriv = props.Invertible("Charge density accumulation rate (C/(s m^3))")

    def __init__(
        self, mesh, survey=None, sigma=None, rho=None, q=None, qMap=None, **kwargs
    ):
        # These below checks can be commented out, correspondingly do
        # not set sigmaMap and rhoMap to None on the super call, to enable
        # derivatives with respect to resistivity/conductivity.
        if sigma is None:
            if rho is None:
                raise ValueError("Must set either conductivity or resistivity.")
        else:
            if rho is not None:
                raise ValueError("Cannot set both conductivity and resistivity.")
        super().__init__(
            mesh=mesh,
            survey=survey,
            sigma=sigma,
            rho=rho,
            sigmaMap=None,
            rhoMap=None,
            **kwargs
        )
        self.q = q
        self.qMap = qMap

    def getRHS(self):
        return self.Vol @ self.q

    def getRHSDeriv(self, source, v, adjoint=False):
        if adjoint:
            return self.qDeriv.T @ (self.Vol @ v)
        return self.Vol @ (self.qDeriv @ v)

    @property
    def deleteTheseOnModelUpdate(self):
        # When enabling resistivity derivatives, uncomment these lines
        # if self.rhoMap is not None:
        #     return super().deleteTheseOnModelUpdate
        if self.storeJ and self.qMap is not None and not self.qMap.is_linear:
            return ["_Jmatrix", "_jtjdiag"]
        return []


class CurrentDensityMap(maps.LinearMap):
    r"""Maps current density to charge density accumulation rate.

    Parameters
    ----------
    mesh : discretize.base.BaseMesh
    active_cells : index_array, optional
        Defaults to all cells being active. This is used to apply a
        0 Nuemann boundary condition at topographic faces.

    Notes
    -----
    This map relates current density :math:`\vec{j}_s` in `A/m^2` to the
    charge density accumulation rate :math:`q` (`C/(s m^3)`) by taking
    the negative divergence.

    .. math::

        q = -\nabla \cdot \vec{j}_s
    """

    def __init__(self, mesh, active_cells=None):
        cv_to_f = mesh.average_cell_vector_to_face
        if active_cells is not None:
            active_cells = validate_active_indices(
                "active_cells", active_cells, mesh.n_cells
            )
            active_cell_comps = np.concatenate(mesh.dim * [active_cells])
            # need to repeat above for each vector component for cells
            # Test which faces are on the boundary of the active domain
            boundary_faces = (mesh.face_divergence.T @ active_cells) != 0
            # use to set the value of j on the boundaries to zero
            # do not need to worry about the faces that are interior to
            # the inactive cells because the values on each side of a
            # face are already set to zero from the projection from active
            # to full space.
            face_weights = sp.diags(1.0 * (~boundary_faces))
            cv_to_f = face_weights @ cv_to_f[:, active_cell_comps]
        A = -mesh.face_divergence @ cv_to_f
        super().__init__(A)


class HydraulicHeadMap(maps.LinearMap):
    r"""Maps hydraulic head to charge density accumulation rate.

    Parameters
    ----------
    mesh : discretize.base.BaseMesh
    L : float or (mesh.n_cells,) array_like
        Cross coupling property model (`A/m^2`).

    Notes
    -----
    This map relates hydraulic head, :math:`h`, in `m` to the volumetric current
    density, :math:`q`, (`A/m^3`) using a cross coupling term to convert the
    negative gradient of hydrualic head into secondary current densities.
    This cross coupling parameter :math:`L` is a multiplication of the hydraulic
    conductivity (m/s) of the material and the ionic density (C/m^3)
    of the fluid.

    The mapping is:

    .. math::

        q = \nabla \cdot L \nabla h
    """

    def __init__(self, mesh, L):
        div = mesh.face_divergence
        MfLiI = mesh.get_face_inner_product(L, invert_model=True, invert_matrix=True)
        A = div @ MfLiI @ div.T @ sp.diags(mesh.cell_volumes, format="csr")
        super().__init__(A)


class Survey(dc.Survey):
    """Streaming potential survey

    Parameters
    ----------
    source_list : list of sources.StreamingCurrents
    """

    @property
    def source_list(self):
        """List of sources.

        Returns
        -------
        list of sources.StreamingCurrents
        """
        return self._source_list

    @source_list.setter
    def source_list(self, new_list):
        self._source_list = validate_list_of_types(
            "source_list",
            new_list,
            StreamingCurrents,
        )



import numpy as np
from numpy.lib import recfunctions as recFunc

from ..frequency_domain.survey import Survey
from ...data import Data as BaseData
from ...utils import mkvc
from .sources import PlanewaveXYPrimary
from .receivers import PointNaturalSource, Point3DTipper
from .utils.plot_utils import DataNSEMPlotMethods

#########
# Data
#########


class Data(BaseData, DataNSEMPlotMethods):
    """Data class for NSEMdata.

    Stores the data vector indexed by the survey.

    Parameters
    ----------
    survey : SimPEG.survey.Survey
        Natural source EM survey
    dobs : numpy.ndarray
        Observed data
    relative_error : numpy.ndarray, optional
        Relative error
    noise_floor : numpy.ndarray, optional
        Noise floor
    """

    def __init__(self, survey, dobs=None, relative_error=None, noise_floor=None):
        BaseData.__init__(self, survey, dobs, relative_error, noise_floor)

    def toRecArray(self, returnType="RealImag"):
        """
        Returns a numpy.recarray for a SimpegNSEM impedance data object.

        Parameters
        ----------
        returnType : str, default: "RealImag:
            Switches between returning a receiver array where the impedance is split to real and imaginary ('RealImag') or is a complex ('Complex')

        Returns
        -------
        numpy.recarray
            Record array with data, with indexed columns
        """

        # Define the record fields
        dtRI = [
            ("freq", float),
            ("x", float),
            ("y", float),
            ("z", float),
            ("zxxr", float),
            ("zxxi", float),
            ("zxyr", float),
            ("zxyi", float),
            ("zyxr", float),
            ("zyxi", float),
            ("zyyr", float),
            ("zyyi", float),
            ("tzxr", float),
            ("tzxi", float),
            ("tzyr", float),
            ("tzyi", float),
        ]
        dtCP = [
            ("freq", float),
            ("x", float),
            ("y", float),
            ("z", float),
            ("zxx", complex),
            ("zxy", complex),
            ("zyx", complex),
            ("zyy", complex),
            ("tzx", complex),
            ("tzy", complex),
        ]

        for src in self.survey.source_list:
            # Temp array for all the receivers of the source.
            # Note: needs to be written more generally,
            # using diffterent rxTypes and not all the data at the locations
            # Assume the same locs for all RX
            locs = src.receiver_list[0].locations
            if locs.shape[1] == 1:
                locs = np.hstack((np.array([[0.0, 0.0]]), locs))
            elif locs.shape[1] == 2:
                locs = np.hstack((np.array([[0.0]]), locs))
            tArrRec = np.concatenate(
                (
                    src.frequency * np.ones((locs.shape[0], 1)),
                    locs,
                    np.nan * np.ones((locs.shape[0], 12)),
                ),
                axis=1,
            ).view(dtRI)
            # Get the type and the value for the DataNSEM object as a list
            typeList = [
                [rx.orientation, rx.component, self[src, rx]]
                for rx in src.receiver_list
            ]
            # Insert the values to the temp array
            for k, c, val in typeList:
                zt_type = "t" if "z" in k else "z"
                key = zt_type + k + c[0]
                tArrRec[key] = mkvc(val, 2)
            # Masked array

            try:
                outTemp = recFunc.stack_arrays((outTemp, tArrRec))
            except NameError:
                outTemp = tArrRec.copy()

            if "RealImag" in returnType:
                outArr = outTemp.copy()
            elif "Complex" in returnType:
                # Add the real and imaginary to a complex number
                outArr = np.empty(outTemp.shape, dtype=dtCP)
                for comp in ["freq", "x", "y", "z"]:
                    outArr[comp] = outTemp[comp].copy()
                for comp in ["zxx", "zxy", "zyx", "zyy", "tzx", "tzy"]:
                    outArr[comp] = (
                        outTemp[comp + "r"].copy() + 1j * outTemp[comp + "i"].copy()
                    )
            else:
                raise NotImplementedError(
                    "{:s} is not implemented, as to be RealImag or Complex."
                )

        # Return
        return outArr

    @classmethod
    def fromRecArray(cls, recArray, srcType="primary"):
        """Class method that reads in a numpy record array to NSEMdata object.

        Parameters
        ----------
        recArray : numpy.ndarray
            Record array with the data. Has to have ('freq','x','y','z') columns and some ('zxx','zxy','zyx','zyy','tzx','tzy')
        srcType : str, default: "primary"
            The type of SimPEG.EM.NSEM.SrcNSEM to be used. Either "primary" or "total"

        Returns
        -------
        SimPEG.electromagnetics.natural_source.sources.SrcNSEM
            Natural source
        """
        if srcType == "primary":
            src = PlanewaveXYPrimary
        elif srcType == "total":
            src = Planewave_xy_1DhomotD
        else:
            raise NotImplementedError("{:s} is not a valid source type for NSEMdata")

        # Find all the frequencies in recArray
        uniFreq = np.unique(recArray["freq"].copy())
        source_list = []
        dataList = []
        for freq in uniFreq:
            # Initiate receiver_list
            receiver_list = []
            # Find that data for freq
            dFreq = recArray[recArray["freq"] == freq].copy()
            # Find the impedance rxTypes in the recArray.
            rxTypes = [
                comp
                for comp in recArray.dtype.names
                if (len(comp) == 4 or len(comp) == 3) and "z" in comp
            ]
            for rxType in rxTypes:
                # Find index of not nan values in rxType
                notNaNind = ~np.isnan(dFreq[rxType].copy())
                if np.any(notNaNind):  # Make sure that there is any data to add.
                    locs = _rec_to_ndarr(dFreq[["x", "y", "z"]][notNaNind].copy())
                    if dFreq[rxType].dtype.name in "complex128":
                        if "t" in rxType:
                            receiver_list.append(
                                Point3DTipper(locs, rxType[1:3], "real")
                            )
                            dataList.append(dFreq[rxType][notNaNind].real.copy())
                            receiver_list.append(
                                Point3DTipper(locs, rxType[1:3], "imag")
                            )
                            dataList.append(dFreq[rxType][notNaNind].imag.copy())
                        elif "z" in rxType:
                            receiver_list.append(
                                PointNaturalSource(locs, rxType[1:3], "real")
                            )
                            dataList.append(dFreq[rxType][notNaNind].real.copy())
                            receiver_list.append(
                                PointNaturalSource(locs, rxType[1:3], "imag")
                            )
                            dataList.append(dFreq[rxType][notNaNind].imag.copy())
                    else:
                        component = "real" if "r" in rxType else "imag"
                        if "z" in rxType:
                            receiver_list.append(
                                PointNaturalSource(locs, rxType[1:3], component)
                            )
                            dataList.append(dFreq[rxType][notNaNind].copy())
                        if "t" in rxType:
                            receiver_list.append(
                                Point3DTipper(locs, rxType[1:3], component)
                            )
                            dataList.append(dFreq[rxType][notNaNind].copy())

            source_list.append(src(receiver_list, freq))

        # Make a survey
        survey = Survey(source_list)
        dataVec = np.hstack(dataList)
        return cls(survey, dataVec)


def _rec_to_ndarr(rec_arr, data_type=float):
    """
    Function to transform a numpy record array to a nd array.
    dupe of SimPEG.electromagnetics.natural_source.utils.rec_to_ndarr to avoid circular import
    """
    # fix for numpy >= 1.16.0
    # https://numpy.org/devdocs/release/1.16.0-notes.html#multi-field-views-return-a-view-instead-of-a-copy
    return np.array(
        recFunc.structured_to_unstructured(
            recFunc.repack_fields(rec_arr[list(rec_arr.dtype.names)])
        ),
        dtype=data_type,
    )
